import seaborn as sns
from rnavigate import plots, styles


class Profile(plots.Plot):
    """Plot per-nucleotide measurements as colored bars.

    Parameters
    ----------
    num_samples : int
        Number of samples to plot.
    nt_length : int
        Length of the nucleotide sequence.
    region : tuple of int, optional
        Region of the nucleotide sequence to plot.
    **kwargs
        Keyword arguments passed to the plot function.

    Attributes
    ----------
    fig : matplotlib.figure.Figure
        Figure object.
    axes : numpy.ndarray of matplotlib.axes.Axes
        Array of axes objects.
    i : int
        Index of the current plot.
    """

    def __init__(self, num_samples, nt_length, region="all", **kwargs):
        """Initialize the plot."""
        if region == "all":
            self.nt_length = nt_length
            self.region = (1, nt_length)
        else:
            self.nt_length = region[1] - region[0] + 1
            self.region = region
        super().__init__(num_samples, sharey=True, **kwargs)
        self.track_height = 0

    def get_rows_columns(self, rows=None, cols=None):
        """Get the number of rows and columns in the plot.

        Parameters
        ----------
        rows : int, optional
            Number of rows in the plot.
        cols : int, optional
            Number of columns in the plot. This is ignored and set to 1.

        Returns
        -------
        rows : int
            Number of rows in the plot.
        cols : int
            Number of columns in the plot.
        """
        return super().get_rows_columns(rows=rows, cols=1)

    def set_labels(
        self,
        ax,
        axis_title="Reactivity Profile",
        xlabel="Nucleotide Position",
        ylabel="Reactivity",
    ):
        """Set the labels of the plot.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes object.
        axis_title : str, optional
            Title of the axis.
        xlabel : str, optional
            Label of the x-axis.
        ylabel : str, optional
            Label of the y-axis.
        """
        ax.set_title(axis_title, loc="left")
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

    def plot_data(
        self,
        profile,
        annotations,
        domains,
        label,
        plot_error=True,
        column=None,
        seqbar=True,
        annotations_mode="track",
        nt_ticks=(20, 5),
    ):
        """Plot data to the current (or specified) axes.

        Parameters
        ----------
        profile : rnavigate.profile.Profile
            Profile object.
        annotations : list of rnavigate.data.Annotation
            List of annotation objects to plot along the sequence.
        domains : list of rnavigate.data.Annotation, optional
            List of domains to plot along the sequence.
        label : str
            label for the y-axis.
        plot_error : bool, optional
            Whether to plot the error bars.
        column : str, optional
            Column of `profile` to plot.
        seqbar : bool, optional
            Whether to plot the sequence track.
        annotations_mode : "track" or "bar", defaults to "track"
            Mode of the annotations track.
        nt_ticks : tuple of int, optional
            Major and minor tick interval for the nucleotide axis.
        """
        ax = self.get_ax()
        if column is not None:
            profile.metric = column
        column = profile.metric
        plots.plot_profile_bars(
            ax=ax, profile=profile, plot_error=plot_error, region=self.region
        )
        self.add_colorbar_args(profile.cmap)

        track_unit = 0.03
        annotations_track = 2 * track_unit * len(annotations)
        annotations_track *= annotations_mode == "track"
        domains_track = 3 * track_unit * (domains is not None)
        sequence_track = track_unit * seqbar
        self.track_height = sequence_track + annotations_track + domains_track
        if seqbar:
            plots.plot_sequence_track(
                ax=ax,
                sequence=profile.sequence,
                yvalue=self.track_height * -1,
                height=sequence_track,
                region=self.region,
                ytrans="axes",
            )
            self.add_colorbar_args(styles.get_nt_cmap())
        if domains is not None:
            for domain in domains:
                plots.plot_domain_track(
                    ax=ax,
                    spans=domain,
                    yvalue=sequence_track - self.track_height,
                    height=domains_track,
                    region=self.region,
                    ytrans="axes",
                )
        for i, annotation in enumerate(annotations):
            plots.plot_annotation_track(
                ax=ax,
                annotation=annotation,
                mode=annotations_mode,
                yvalue=(
                    sequence_track
                    + domains_track
                    - self.track_height
                    + (i + 0.5) * (4 * track_unit)
                ),
                height=2 * track_unit,
                ytrans="axes",
                region=self.region,
            )

        self.i += 1
        ylabel = column.replace("_", " ")
        self.set_labels(ax=ax, ylabel=ylabel, axis_title=label)
        self.set_axis(ax=ax, sequence=profile.sequence, nt_ticks=nt_ticks)

    def set_figure_size(
        self,
        height_ax_rel=None,
        width_ax_rel=0.03,
        width_ax_in=None,
        height_ax_in=2,
        height_gap_in=1,
        width_gap_in=0.5,
        top_in=1,
        bottom_in=1,
        left_in=1,
        right_in=1,
    ):
        """Set the size of the figure.

        Parameters
        ----------
        height_ax_rel : float, optional
            Height of the axes relative to the y-axis limits.
        width_ax_rel : float, defaults to 0.03
            Width of the axes relative to the x-axis limits.
        width_ax_in : float, optional
            Width of the axes in inches.
        height_ax_in : float, defaults to 2
            Height of the axes in inches.
        height_gap_in : float, defaults to 1
            Height of the gap between axes in inches.
        width_gap_in : float, defaults to 0.5
            Width of the gap between axes in inches.
        top_in : float, defaults to 1
            Height of the top margin in inches.
        bottom_in : float, defaults to 1
            Height of the bottom margin in inches.
        left_in : float, defaults to 1
            Width of the left margin in inches.
        right_in : float, defaults to 1
            Width of the right margin in inches.
        """
        super().set_figure_size(
            height_ax_rel=height_ax_rel,
            width_ax_rel=width_ax_rel,
            width_ax_in=width_ax_in,
            height_ax_in=height_ax_in,
            height_gap_in=height_gap_in + 2 * self.track_height,
            width_gap_in=width_gap_in,
            top_in=top_in,
            bottom_in=bottom_in + 2 * self.track_height,
            left_in=left_in,
            right_in=right_in,
        )

    def set_axis(self, ax, sequence, nt_ticks=(20, 5)):
        """Set up axis properties for aesthetics.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes object.
        sequence : str
            Nucleotide sequence.
        nt_ticks : tuple of int, optional
            Major and minor tick interval for the nucleotide axis.
        """
        xlim = self.region
        ax.set_xlim([xlim[0] - 0.5, xlim[1] + 0.5])
        plots.set_nt_ticks(
            ax=ax,
            sequence=sequence,
            region=self.region,
            major=nt_ticks[0],
            minor=nt_ticks[1],
        )
        ax.spines["bottom"].set(
            position=("axes", self.track_height * -1), visible=False
        )
        ax.spines["left"].set_visible(False)
        ax.grid(axis="y", visible=True)
        ax.tick_params(axis="y", length=0, grid_alpha=0.4)
        sns.despine(ax=ax, bottom=True, left=True)



"""Parsing function for rnavigate.Sample data_keywords."""

import os
from rnavigate import data


__all__ = ["data_keyword_defaults", "create_data", "get_sequence"]

_required = object()

data_keyword_defaults = {
    "sequence": {"data_class": data.Sequence},
    "shapemap": {"data_class": data.SHAPEMaP},
    "shapemap_rnaframework": {"data_class": data.SHAPEMaP.from_rnaframework},
    "dmsmap": {"data_class": data.SHAPEMaP, "normalize": "DMS"},
    "dancemap": {"data_class": data.DanceMaP, "component": _required},
    "rnpmap": {"data_class": data.RNPMaP},
    "ringmap": {"data_class": data.RINGMaP, "sequence": "default_profile"},
    "pairmap": {"data_class": data.PAIRMaP, "sequence": "default_profile"},
    "shapejump": {"data_class": data.SHAPEJuMP, "sequence": _required},
    "pairprob": {"data_class": data.PairingProbability, "sequence": "default_profile"},
    "ss": {"data_class": data.SecondaryStructure},
    "ss_pairs": {
        "data_class": data.SecondaryStructure.from_pairs_list,
        "sequence": _required,
    },
    "pdb": {"data_class": data.PDB, "chain": _required},
    "allpossible": {"data_class": data.AllPossible, "sequence": _required},
    "motif": {
        "data_class": data.Motif,
        "name": _required,
        "sequence": _required,
        "color": _required,
    },
    "orfs": {
        "data_class": data.ORFs,
        "name": _required,
        "sequence": _required,
        "color": _required,
    },
    "spans": {
        "data_class": data.Annotation,
        "name": _required,
        "sequence": _required,
        "color": _required,
    },
    "sites": {
        "data_class": data.Annotation,
        "name": _required,
        "sequence": _required,
        "color": _required,
    },
    "group": {
        "data_class": data.Annotation,
        "name": _required,
        "sequence": _required,
        "color": _required,
    },
    "primers": {
        "data_class": data.Annotation,
        "name": _required,
        "sequence": _required,
        "color": _required,
    },
    "domains": {
        "data_class": data.domains,
        "sequence": _required,
        "colors": _required,
        "names": _required,
    },
}


def create_data(data_keyword, inputs, sample=None):
    """Convenience function for creating rnavigate.data objects. This function
    is used to parse **data_keywords passed to rnavigate.Sample, but can also
    be used on it's own using the same syntax.

    Parameters
    ----------
        data_keyword : string
            a data keyword sting
            if inputs is a dictionary, this argument is ignored
        inputs : any
            If a dictionary, the first key must be the standard data keyword
            Otherwise, the expected type depends on `data_keyword`
        sample : rnavigate.Sample, optional
            Requried only if inputs is a dictionary containing a "sequence" key
            with a value that is an assigned data keyword
                {"sequence": "data_keyword"}
                    is replaced with:
                {"sequence": sample.get_data("data_keyword").sequence}
            Defaults to None.

    Returns
    -------
        rnavigate.data.Sequence object
            the new RNAvigate data object

    Examples
    --------
        In this example we create a Sequence object from a fasta file. There are
        two valid syntaxes:

            create_data(
                data_keyword="sequence",
                inputs="my_sequence.fa"
                )

        OR

            create_data(
                data_keyword="arbitrary_string",
                inputs={"sequence": "my_sequence.fa"}
                )

        Returns: rnavigate.data.Sequence(input_data="my_sequence.fa")
    """
    if sample is not None:
        label = f"{sample.sample} ({data_keyword})"
    else:
        label = data_keyword
    # if given existing rnavigate.data object return
    if isinstance(inputs, data.Sequence):
        return inputs

    # data_keyword="A", inputs="B"
    # OR
    # data_keyword="arbitrary", inputs={"A": "B", etc.}
    # becomes
    # data_keyword="A", inputs={"input_data": "B", etc.}
    if isinstance(inputs, dict):
        first_key = list(inputs.keys())[0]
        data_keyword = first_key
        inputs["input_data"] = inputs.pop(first_key)
    else:
        inputs = {"input_data": inputs}

    # handle special cases
    if data_keyword == "allpossible":
        inputs["sequence"] = inputs.pop("input_data")
    elif data_keyword in ["sites", "spans", "primers", "group"]:
        inputs["annotation_type"] = data_keyword

    # retrieve defaults for given data_class
    try:
        inputs = data_keyword_defaults[data_keyword] | inputs
        data_constructor = inputs.pop("data_class")
    except KeyError as exception:
        raise KeyError(f"{data_keyword} is not a valid data keyword.") from exception

    # get sequence from another object if appropriate
    if "sequence" in inputs:
        inputs["sequence"] = get_sequence(inputs["sequence"], sample)

    # check if any required arguments are not provided
    required_arguments = []
    for key, value in inputs.items():
        if value is _required:
            required_arguments.append(key)
    if len(required_arguments) > 0:
        raise ValueError(
            f"{data_keyword} missing required arguments:\n"
            ", ".join(required_arguments)
        )

    # instantiate and return the data class
    try:
        data_obj = data_constructor(**inputs)
        if isinstance(data_obj, data.Sequence) and data_obj.name is None:
            data_obj.name = label
        return data_obj
    except BaseException as exception:
        raise ValueError(f"data_keyword={data_keyword}\ninputs={inputs}") from exception


def get_sequence(sequence, sample=None, default=None):
    """Flexible function that returns a Data object containing a sequence.

    Parameters
    ----------
        sequence : any
            a sequence string, a data keyword string, or a data object
        sample : rnavigate.Sample, optional
            A sample is required if `sequence` is a data keyword string
            Defaults to None
        default : any, optional
            same format as sequence above. This value is used if `sequence` is None
            Defaults to None.

    Returns
    -------
        rnavigate.Sequence
            a sequence object containing the desired sequence
    """
    if sequence is None and default is None:
        raise ValueError("A sequence must be provided.")
    elif sequence is None:
        sequence = default
    try:
        sequence = sample.get_data(sequence)
    except ValueError:
        pass
    if isinstance(sequence, data.Sequence):
        return sequence
    if isinstance(sequence, str) and os.path.isfile(sequence):
        return data.Sequence(sequence)
    if isinstance(sequence, str) and all([nt.upper() in "AUCGT.-" for nt in sequence]):
        return data.Sequence(sequence)
    else:
        raise ValueError(f"Cannot find sequence from ({sequence})")



#!/usr/bin/env python3
# -----------------------------------------------------
# Utilities functions for fpocket-R
# Seth Veenbaas
# Weeks Lab, UNC-CH
# 2024
#
# Version 1.3.0
#
# -----------------------------------------------------
import os
import re
from glob import glob
from prody import *
import numpy as np


def fetch_pdb(pdb_id : str) -> str:
    ''' Downloads structure from the PDB.
    Fetches .pdb format or .cif format if the .pdb is unavailable.
    Args:
        pdb_id: 4-character alphanumeric PDB identifier
    Returns:
        str: path to .pdb (or .cif) file
    '''
    pdb_id_lower = pdb_id.lower()
    pdb_filename = fetchPDB(f'{pdb_id_lower}', compressed=False, quiet=False)
    return pdb_filename

def natsorted(filenames : list) -> list:
    """ Sorts filenames by digits in the filenames.
    Supports multiple numbers in a filename.
    Natsorted example: [
        'pocket1_atm.pdb',
        'pocket2_atm.pdb',
        '2pocket11_atm.pdb',
        'pocket10_atm.pdb',
        'pocket11_atm.pdb',
        'pocket11_atm5.pdb',
        'pocket20_atm.pdb'
    ]
    Args:
        filenames (list): List of strings to be sorted by digits.
    Returns:
        list: Sorted list, ordered by digits.
    """    
    def extract_sort_key(filename):
        # Extract all numbers in the filename as integers, or return the filename as-is for non-numeric parts
        return [int(num) if num.isdigit() else num.lower() 
                for num in re.split(r'(\d+)', filename)]
    
    return sorted(filenames, key=extract_sort_key)


def is_accessible(path : str, file_name : str) -> None:
    """Checks if input file is accessible.
       Stops program and displays:
        - FileNotFoundError: if file is not accessible/found.
        - TypeError: if path argument for input file is not a string.

    Args:
        path (str): path to file being checked.
        file_name (str): name of file being checked for display in error.

    Raises:
        FileNotFoundError: Not able to read input file.
        PermissionError: Not able to access {file_name}.')
    """
    if not os.path.exists(path):
        print(f'FileNotFoundError: The file {file_name} does not exist: {path}')
        raise FileNotFoundError(f'The file {file_name} does not exist at: {path}.')


    if not os.access(path, os.R_OK):
        print(f'PermissionError: Not able to access {file_name}.')
        raise PermissionError(f'Not able to access {file_name}.')


def get_first_rna_chain(pdb : str) -> str:
    """Identifies first RNA chain in .pdb/.cif file.

    Args:
        pdb (str): Path to input .pdb/.cif file.

    Returns:
        str: Chain identifier of the first chain containing RNA.
    """

    structure = prody.parsePDB(pdb)
        
    for ch in structure.getHierView():
        chid = ch.getChid()
        if structure.select(f'chain {chid} and nucleotide'):
            print(f'\nAutomatically selecting chain: {chid}\n'
                  'Use --chain to manually specify an RNA chain.\n')
            return chid
        
    print(f'No RNA chain found in {pdb}.\n'
                    'Verify that your pdb structure contains RNA.\n'
                    'Or manually set an RNA chain with (--chain) option.')
    raise ValueError(f'No RNA chain found in {pdb}.\n'
                    'Verify that your pdb structure contains RNA.\n'
                    'Or manually set an RNA chain with (--chain) option.')


def is_rna_chain(pdb : str, chain : str) -> None:
    """Identifies if a chain an a .pdb/.cif file contains rna.

    Args:
        pdb (str): Path to input .pdb/.cif file.
        chain (str): Chain identifier for RNA chain.

    Returns:
        None
    """
    structure = prody.parsePDB(pdb)
    chains = chain.split(',')
    hv = structure.getHierView()
    chids = list(hv)

    for chain in chains:
        if not structure.select(f'chain {chain}'):
            print(f'KeyError: The input {pdb} does not contain the chain: {chain}.\n'
            'Input a valid chain id using the (--chain) option.\n'
            'NOTE: Chain identifiers are case sensitive.\n'
            f'Chain(s) present in {pdb}:\n{chids}')
            raise KeyError(f'The input {pdb} does not contain the chain: {chain}.'
            'Input a valid chain id using the (--chain) option.'
            'NOTE: Chain identifiers are case sensitive.')
        elif not structure.select(f'chain {chain} nucleic'):
            print(f'KeyError: Chain {chain} of {pdb} does not contain rna.\n'
            'Input a valid RNA chain id using the (--chain) option.\n'
            'NOTE: Chain identifiers are case sensitive.'
            f'Chain(s) present in {pdb}:\n{chids}')
            raise KeyError(f'Chain {chain} of {pdb} does not contain rna.'
            'Input a valid RNA chain id using the (--chain) option.\n'
            'NOTE: Chain identifiers are case sensitive.')


def get_file_paths(
    analysis : str, name :str , pdb : str, state : int
    )-> tuple[str, str, str, str, list[str], str, str]:
    """Gets paths to required input files and check if they are accessible.

    Args:
        analysis (str): path to directory containing fpocket outputs.
        name (str): user specified filename prefix for analysis outputs.
        pdb (str): Path to input .pdb file.
        state (int): Structural state to analyze.

    Returns:
        str: valid path to *.pdb file
        str: valid path to fpocket generated *_out.pdb file
        str: valid path to fpocket generated *_out.pqr file
        str: valid path to fpocket generated *_info.txt file
        str: valid paths to pockets/pocket*_atm.pdb files
        str: 4 character pdb code
        str: filename prefix for analysis and figure output files
    """
    cwd = os.getcwd()
    analysis_basename = os.path.basename(analysis)[0:-4]

    if not pdb:
        pdb_basename = analysis_basename.replace('_clean', '')
        pdb = os.path.join(cwd, f'{pdb_basename}.pdb')
        is_accessible(pdb, 'pdb')

    pdb_out = os.path.join(cwd, analysis,
                           f'{analysis_basename}_out.pdb')
    is_accessible(pdb_out, 'pdb_out')

    pqr_out = os.path.join(cwd, analysis,
                           f'{analysis_basename}_pockets.pqr')
    is_accessible(pqr_out, 'pqr_out')

    info_txt = os.path.join(cwd, analysis,
                            f'{analysis_basename}_info.txt')
    is_accessible(info_txt, 'info_txt')

    pockets_dir = os.path.join(cwd, analysis, 'pockets')
    
    # Sort pockets_out filenames naturally (eg: 1,2,3...10,11....20,21)
    pockets_out = natsorted(glob(f'{pockets_dir}/pocket*_atm.pdb'))

    # Get PDB identifiers from the path to the fpocket out.pdb.
    pdb_basename = os.path.basename(pdb)
    pdb_code = pdb_basename[0:4]

    if state:
        name = f'{name}_state{state}'

    return pdb, pdb_out, pqr_out, info_txt, pockets_out, pdb_code, name


def calc_npr(atom_group: object) -> tuple[float, float]:
    """Generates the normalized PMI ratio for an input object.

    Args:
        atomgroup (object): prody atomgroup

    Returns:
        float: normalized principle ratio 1 = I1/I3
        float: normalized principle ratio 2 = I2/I3
    """
    # calculates inertia tensor for atomgroup
    inertia_tensor = calc_inertia_tensor(atom_group)

    # calculates the principle moments of inertia for the atomgroup
    pmi = calc_pmi(inertia_tensor)
    I1 = pmi[0]
    I2 = pmi[1]
    I3 = pmi[2]

    # calcualtes normalize PMI ratios
    npr1 = I1 / I3
    npr2 = I2 / I3

    return npr1, npr2


def calc_inertia_tensor(atom_group: object) -> np.array:
    """Generates an interia tensor for a prody atomgroup object.

    Args:
        atomgroup (object): prody atomgroup

    Returns:
        np.array: 3x3 inertia tensor
    """
    # Calculate center of mass
    totmass = 0.0
    x_com, y_com, z_com = 0, 0, 0

    for atom in atom_group:

        atom_coords = atom.getCoords()

        if atom.getResname() == 'STP':
            atom_mass = 1
        else:
            atom_mass = atom.getMass()

        x_com += atom_coords[0] * atom_mass
        y_com += atom_coords[1] * atom_mass
        z_com += atom_coords[2] * atom_mass
        totmass += atom_mass

    x_com /= totmass
    y_com /= totmass
    z_com /= totmass

    # Make and populate inertia tensor

    I = []
    for _ in range(9):
        I.append(0)

    for atom in atom_group:
        atom_coords = atom.getCoords()
        if atom.getResname() == 'STP':
            atom_mass = 1
        else:
            atom_mass = atom.getMass()

        temp_x, temp_y, temp_z = atom_coords[0], atom_coords[1], atom_coords[2]
        temp_x -= x_com
        temp_y -= y_com
        temp_z -= z_com

        I[0] += atom_mass * (temp_y ** 2 + temp_z ** 2)
        I[4] += atom_mass * (temp_x ** 2 + temp_z ** 2)
        I[8] += atom_mass * (temp_x ** 2 + temp_y ** 2)
        I[1] -= atom_mass * temp_x * temp_y
        I[3] -= atom_mass * temp_x * temp_y
        I[2] -= atom_mass * temp_x * temp_z
        I[6] -= atom_mass * temp_x * temp_z
        I[5] -= atom_mass * temp_y * temp_z
        I[7] -= atom_mass * temp_y * temp_z

    # generate inertia tensor
    inertia_tensor = np.array([(I[0:3]), (I[3:6]), (I[6:9])])

    return inertia_tensor


def calc_pmi(inertia_tensor : np.array) -> list:
    """Calaculates sorted eigen values (principle moments of inertia) 
    for an imput inertia tensor.

    Args:
        inertia_tensor (array): 3x3 inertia tensor

    Returns:
        list: sorted principle moments of inertia (PMI) [I1, I2, I3]
    """
    # calculate eigen values and eigen vectors
    eigvals, _ = np.linalg.eig(inertia_tensor)

    # sort eigen values > (I1, I2, I3)
    eig_ord = np.argsort(eigvals)

    sorted_eigvals = eigvals[eig_ord]

    return sorted_eigvals


def get_offset(pdb : str, chain : str, offset : int) -> int:
    """Calculates offset (or difference) between the nucleotide index (start at 1)
    and residue number for the first nucleotide in the first chain.

    Args:
        pdb (str): Path to input .pdb file.
        chain (str): Chain identifier for RNA chain.
        offset (int): Sequence offset between .pdb and .nsd file (default=None).

    Returns:
        int: Nucleotide offset of PDB chain.
    """

    if ',' in chain:
        chain = chain.split(',')[0]

    if pdb.endswith('.pdb'):
        structure = parsePDB(pdb)
    elif pdb.endswith('.cif'):
        structure = parsePDB(pdb)
    else:
        print(f'Could not parse structure: {pdb}')
        return None

    if structure:
        hv = structure.getHierView()
        ch = hv[chain]

    else:
        print(f'Could not parse structure: {pdb}')
        return None
    
    if ch:
        resn = ch.getResnums().min()
        offset = resn - 1
    
    else:
        print(f'Chain {chain} not found in structure: {pdb}')
        return None
    
    return offset

# Function to get the last processed state
def get_last_processed_state(state_tracker_filename):
    if os.path.exists(state_tracker_filename):
        with open(state_tracker_filename, 'r') as f:
            return int(f.read().strip())  # Read the last state
    else:
        return 0  # If no tracker file, start from state 1

# Function to update the state tracker
def update_last_processed_state(state_tracker_filename, state):
    with open(state_tracker_filename, 'w') as f:
        f.write(str(state))  # Write the current state to the file
        


#!/usr/bin/env python
"""DeltaSHAPE for detecting meaningful changes in SHAPE reactivity between two samples.

Parameters are optimized for detecting in cell vs. cell free protein protections and
enhancements, but useful for identifying any useful differences.

Copyright Matthew J. Smola 2015
Largely rewritten for RNAvigate by Patrick Irving 2023
"""

###########################################################################
# GPL statement:                                                          #
#                                                                         #
# This program is free software: you can redistribute it and/or modify    #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# This program is distributed in the hope that it will be useful,         #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.   #
###########################################################################

from rnavigate import data, plots, Sample
from scipy.stats import zscore


class DeltaSHAPE(Sample):
    """Detects meaningful differences in chemical probing reactivity

    Citation: (doi:10.1021/acs.biochem.5b00977)

    Algorithm:
        1. Extract SHAPE-MaP sequence, normalized profile, and normalized
           standard error from given samples
        2. Calculated smoothed profiles (mean) and propagate standard errors
           over rolling windows
        3. Subtract raw and smoothed normalized profiles and propogate errors
        4. Calculate Z-factors for smoothed data. This is the magnitude of the
           difference relative to the standard error
        5. Calculate Z-scores for smoothed data. This is the magnitude of the
           difference in standard deviations from the mean difference
        6. Call sites. Called sites must have # nucleotides that pass Z-factor
           and Z-score thresholds per window.

    Smoothing window size, Z factor threshold, Z score threshold, site-calling
    window size and minimum nucleotides per site can be specified.
    """

    def __init__(
        self,
        sample1,
        sample2,
        profile="shapemap",
        smoothing_window=3,
        zf_coeff=1.96,
        ss_thresh=1,
        site_window=3,
        site_nts=2,
    ):
        """Performs DeltaSHAPE analysis between samples 1 and 2

        Required Arguments:
            sample1 (rnavigate.Sample)
                First sample to compare
            sample2 (rnavigate.Sample)
                Second sample to compare

        Optional Arguments:
            profile (string)
                Data keyword pointing to SHAPE-MaP data in samples 1 and 2
                Defaults to "shapemap"
            smoothing_window (integer)
                Size of windows for data smoothing
                Defaults to 3
            zf_coeff (float)
                Sites must have a difference more than zf_coeff standard errors
                Defaults to 1.96 (95% confidence interval)
            ss_thresh (float)
                Sites must have a difference that is ss_thresh standard
                deviations from the mean difference
                Defaults to 1
            site_window (integer)
                Number of nucleotides to include when calling sites
                Defaults to 3
            site_nts (integer)
                Number of nts within site_window that must pass thresholds
                Defaults to 2
        """
        self.parameters = {}
        profile_1 = sample1.get_data(profile)
        profile_1 = profile_1.get_aligned_data(
            data.SequenceAlignment(profile_1, profile_1)
        )
        profile_2 = sample2.get_data(profile)
        profile_2 = profile_2.get_aligned_data(
            data.SequenceAlignment(profile_2, profile_1)
        )
        super().__init__(
            sample=f"{sample1.sample} vs. {sample2.sample}",
            inherit=[sample1, sample2],
            deltashape=DeltaSHAPEProfile((profile_1, profile_2)),
            profile_1=profile_1,
            profile_2=profile_2,
        )
        self.calculate_deltashape(
            smoothing_window=smoothing_window,
            zf_coeff=zf_coeff,
            ss_thresh=ss_thresh,
            site_window=site_window,
            site_nts=site_nts,
        )

    def calculate_deltashape(
        self, smoothing_window=3, zf_coeff=1.96, ss_thresh=1, site_window=2, site_nts=3
    ):
        """Calculate or recalculate deltaSHAPE profile and called sites

        Optional Arguments:
            smoothing_window (integer)
                Size of windows for data smoothing
                Defaults to 3
            zf_coeff (float)
                Sites must have a difference more than zf_coeff standard errors
                Defaults to 1.96 (95% confidence interval)
            ss_thresh (float)
                Sites must have a difference that is ss_thresh standard
                deviations from the mean difference
                Defaults to 1
            site_window (integer)
                Number of nucleotides to include when calling sites
                Defaults to 3
            site_nts (integer)
                Number of nts within site_window that must pass thresholds
                Defaults to 2
        """
        self.parameters = {
            "smoothing_window": smoothing_window,
            "zf_coeff": zf_coeff,
            "ss_thresh": ss_thresh,
            "site_nts": site_nts,
            "site_window": site_window,
        }
        deltashape = self.data["deltashape"]
        deltashape.calculate_deltashape(
            smoothing_window=smoothing_window,
            zf_coeff=zf_coeff,
            ss_thresh=ss_thresh,
            site_window=site_window,
            site_nts=site_nts,
        )
        self.set_data("protections", deltashape.get_protections_annotation())
        self.set_data("enhancements", deltashape.get_enhancements_annotation())

    def plot(self, region="all"):
        """Plot the deltaSHAPE result

        Optional arguments:
            region (list of 2 integers)
                start and end positions to plot
                Defaults to "all".

        Returns:
            rnav.plots.Profile: The plot object
        """
        plot = plots.Profile(1, self.data["deltashape"].length, region=region)
        plot.plot_data(
            profile=self.data["deltashape"],
            annotations=self.get_data(["protections", "enhancements"]),
            domains=None,
            plot_error=True,
            label=self.sample,
        )
        plot.set_figure_size()
        return plot


class DeltaSHAPEProfile(data.Profile):
    """Profile data class for performing deltaSHAPE analysis"""

    def __init__(
        self,
        input_data,
        metric="Smooth_diff",
        metric_defaults=None,
        sequence=None,
        name=None,
        **kwargs,
    ):
        """Create the deltaSHAPE Profile

        Args:
            input_data (tuple of RNAvigate Profiles or Pandas Dataframe)
                if tuple of Profiles, the unified Dataframe will be created
        """
        # STEP ONE
        # extract relevant information from sample 1 and 2 profiles
        # sequence, normalized profile, normalized standard error
        columns = ["Nucleotide", "Sequence", "Norm_profile", "Norm_stderr"]
        if isinstance(input_data, (tuple, list)):
            profile1, profile2 = input_data
            df_1 = profile1.data[columns]
            df_2 = profile2.data[columns]
            input_data = df_1.merge(
                df_2, how="left", on=["Nucleotide", "Sequence"], suffixes=("_1", "_2")
            )
        if metric_defaults is None:
            metric_defaults = {}
        metric_defaults = {
            "Smooth_diff": {
                "metric_column": "Smooth_diff",
                "error_column": "Smooth_diff_stderr",
                "color_column": "Class",
                "cmap": ["lightgrey", "#7F3B95", "#3EB452"],
                "normalization": "none",
                "values": None,
                "extend": "neither",
                "title": "deltaSHAPE",
                "ticks": [0, 1, 2],
                "tick_labels": ["other", "protection", "enhancement"],
                "alpha": 0.7,
            },
        } | metric_defaults
        super().__init__(
            input_data=input_data,
            metric=metric,
            metric_defaults=metric_defaults,
            sequence=sequence,
            name=name,
            **kwargs,
        )

    def calculate_deltashape(
        self, smoothing_window=3, zf_coeff=1.96, ss_thresh=1, site_window=3, site_nts=2
    ):
        """Calculate the deltaSHAPE profile metrics

        Args:
            smoothing_window (int, optional): Defaults to 3.
            zf_coeff (float, optional): Defaults to 1.96.
            ss_thresh (int, optional): Defaults to 1.
            site_window (int, optional): Defaults to 3.
            site_nts (int, optional): Defaults to 2.
        """

        # STEP TWO
        # smooth data and errors
        def propagate_errors(errors):
            errors = errors[~errors.isna()]
            return (errors**2).sum() ** 0.5 / len(errors)

        self.calculate_windows(
            column="Norm_profile_1",
            window=smoothing_window,
            new_name="Smooth_profile_1",
            method="mean",
            minimum_points=1,
            mask_na=True,
        )
        self.calculate_windows(
            column="Norm_stderr_1",
            window=smoothing_window,
            new_name="Smooth_stderr_1",
            method=propagate_errors,
            minimum_points=1,
            mask_na=True,
        )
        self.calculate_windows(
            column="Norm_profile_2",
            window=smoothing_window,
            new_name="Smooth_profile_2",
            method="mean",
            minimum_points=1,
            mask_na=True,
        )
        self.calculate_windows(
            column="Norm_stderr_2",
            window=smoothing_window,
            new_name="Smooth_stderr_2",
            method=propagate_errors,
            minimum_points=1,
            mask_na=True,
        )

        # STEP THREE
        # subtract raw and smoothed data
        self.data["Diff_profile"] = self.data.eval("Norm_profile_1 - Norm_profile_2")
        self.data["Diff_stderr"] = self.data.eval("Norm_stderr_1 + Norm_stderr_2")
        self.data["Smooth_diff"] = self.data.eval("Smooth_profile_1 - Smooth_profile_2")
        self.data["Smooth_diff_stderr"] = self.data.eval(
            "Smooth_stderr_1 + Smooth_stderr_2"
        )
        self.data["Positive"] = self.data["Smooth_diff"] > 0

        # STEP FOUR
        # calculate Z-factors from smoothed data and smoothed errs
        confidence_interval = self.data.eval(
            f"Smooth_stderr_1 + Smooth_stderr_2 * {zf_coeff}"
        )
        difference = self.data["Smooth_diff"].abs()
        self.data["Z_factor"] = 1 - (confidence_interval / difference)

        # STEP FIVE
        # calculate Z-scores from difference of smoothed data
        self.data["Z_score"] = abs(zscore(self.data["Smooth_diff"], nan_policy="omit"))

        # STEP SIX
        # identify site_window nt windows where site_nts are significant
        def determine_significance(series):
            df = self.data.loc[series.index, ["Z_factor", "Z_score"]]
            significant = df["Z_factor"] > 0
            significant &= df["Z_score"] > ss_thresh
            return sum(significant) >= site_nts

        windows = self.data["Nucleotide"].rolling(site_window, center=True)
        self.data["Significant"] = windows.apply(determine_significance)
        self.data["Significant"] = self.data["Significant"].fillna(0).astype(bool)
        self.data["Class"] = 0
        self.data.loc[self.data.eval("Significant & ~ Positive"), "Class"] = 1
        self.data.loc[self.data.eval("Significant & Positive"), "Class"] = 2

    def get_protections_annotation(self):
        """Get an annotations object for the significant protections"""
        is_protected = self.data.eval("Significant & ~ Positive")
        return data.Annotation.from_boolean_array(
            values=is_protected,
            window=1,
            annotation_type="spans",
            name="Protections",
            sequence=self.sequence,
            color="#7F3B95",
        )

    def get_enhancements_annotation(self):
        """Get an annotations object for the significant enhancements"""
        is_enhanced = self.data.eval("Significant & Positive")
        return data.Annotation.from_boolean_array(
            values=is_enhanced,
            window=1,
            annotation_type="spans",
            name="Enhancements",
            sequence=self.sequence,
            color="#3EB452",
        )



"""Contains global plot display settings."""
from functools import wraps
import matplotlib as mpl
import seaborn as sns
from rnavigate import data

# STYLE SHEET
###############################################################################

__all__ = [
    "update_copy",
    "Settings",
    "set_defaults",
    "get_nt_color",
    "get_nt_cmap",
    "apply_style",
]

settings = {
    "sequence_bar": "alphabet",  # "bars"
    "sequence_colors": "rnavigate",  # "old"
    "ss": {
        "structure": {
            "linewidth": 1,
            "zorder": 2,
        },
        "interactions": {
            "linewidth": 3,
            "alpha": None,
            "zorder": 15,
        },
        "spans": {
            "linewidth": 10,
            "alpha": 0.4,
            "zorder": 5,
        },
        "sites": {
            "marker": "o",
            "edgecolor": "none",
            "s": 10**2,
            "alpha": 0.4,
            "zorder": 5,
        },
        "basepairs": {"zorder": 0},
        "nucleotides": {
            "marker": "o",
            "edgecolor": "none",
            "s": 5**2,
            "zorder": 10,
        },
        "sequence": {
            "linewidth": 0.3,
            "s": 3**2,
            "zorder": 20,
        },
        "positions": {"zorder": 25},
    },
}


def update_copy(original_settings, user_settings):
    """Recursively updates and returns a copy of og settings with new settings applied.

    Parameters
    ----------
        original_settings (dict)
            a default settings dictionary, usually rnav.settings
        user_settings (dict)
            a dictionary with only the fields from the original_settings that
            are to be changed

    Returns
    -------
        settings dict
            the original_settings dictionary with the new_settings dictionary
            values recursively applied
    """
    new_settings = dict()
    for k, v in original_settings.items():
        if isinstance(v, dict):
            try:
                new_settings[k] = update_copy(v, user_settings[k])
            except KeyError:
                new_settings[k] = v | {}
        else:
            try:
                new_settings[k] = user_settings[k]
            except KeyError:
                new_settings[k] = v
    return new_settings


class Settings(dict):
    """Context manager for temporarily changing global settings.

    Parameters
    ----------
        user_settings : dict
            a dictionary with only the fields from the original_settings that
            are to be changed
    """

    def __init__(self, user_settings):
        self.original_settings = update_copy(settings, {})
        self.user_settings = update_copy(settings, user_settings)

    def __enter__(self):
        settings.update(self.user_settings)

    def __exit__(self, *args, **kwargs):
        settings.update(self.original_settings)


def set_defaults(context="paper", style="ticks", colors="default", dpi=140):
    """Set or reset the major global style settings.

    Parameters
    ----------
        context : str, defaults to "paper"
            Passed to seaborn.set_context
            Defaults to "paper"
        style : str, defaults to "ticks"
            Passed to seaborn.set_style
        colors : str, defaults to "default"
            Passed to seaborn.set_palette
        dpi : int, defaults to 140
            Sets the dots-per-inch for inline and exported images
    """
    if colors == "default":
        colors = [
            "#0092edff",  # Blue
            "#ff8300ff",  # Orange
            "#a100ffff",  # Purple
            "#edc600ff",  # Yellow
            "#ff48e9ff",  # Pink
            "#3fd125ff",  # Green
        ]
    sns.set_context(context)
    sns.set_style(style)
    sns.set_palette(colors)
    mpl.rcParams["font.sans-serif"].insert(0, "Arial")
    mpl.rcParams["figure.dpi"] = dpi
    mpl.rcParams["svg.fonttype"] = "none"


def get_nt_color(nt, colors=None):
    """Get the RNAvigate color for a given nucleotide.

    Invalid nucleotides are set to gray

    Parameters
    ----------
        nt : str
            a nucleotide letter
        colors "rnavigate" or "old", defaults to settings["sequence_colors"]
            "rnavigate" uses blue, light blue, red, light red for "AUCG"
            "old" uses traditional red, yellow, blue, green for "AUCG"

    Returns:
        color : str
            a hex color string
    """
    if colors is None:
        colors = settings["sequence_colors"]
    try:
        return {
            "old": {
                "A": "#f20000",  # red
                "U": "#f28f00",  # yellow
                "G": "#00509d",  # blue
                "C": "#00c200",  # green
            },
            "rnavigate": {
                "A": "#366ef0",  # blue
                "U": "#9bb9ff",  # light blue
                "G": "#f04c4c",  # red
                "C": "#ffa77c",  # light red
            },
        }[colors][nt.upper().replace("T", "U")]
    except KeyError:
        return "#aaaaaa"


def get_nt_cmap(colors=None):
    """Get an rnavigate color map for nucleotides.

    Parameters
    ----------
        colors "rnavigate" or "old", defaults to settings["sequence_colors"]
            "rnavigate" uses blue, light blue, red, light red for "AUCG"
            "old" uses traditional red, yellow, blue, green for "AUCG"

    Returns
    -------
        cmap : rnavigate.data.colors.ScalarMappable (matplotlib.cm.ScalarMappable)
            a color map for nucleotides
    """
    return data.ScalarMappable(
        cmap=[get_nt_color(nt, colors=colors) for nt in "AUGC"],
        normalization="none",
        values=None,
        title="sequence",
        tick_labels=["A", "U", "G", "C"],
    )


def apply_style(style_dict):
    """Decorator for applying matplotlib style settings to a function."""

    def decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            with mpl.rc_context(style_dict):
                return function(*args, **kwargs)

        return wrapper

    return decorator


# TODO: these belong in settings dict
# ShapeMapper Plot Styles
sm = {
    "font.family": "sans-serif",
    "pdf.fonttype": 42,
    # use TrueType fonts when exporting PDFs
    # (embeds most fonts - this is especially
    #  useful when opening in Adobe Illustrator)
    "xtick.direction": "out",
    "ytick.direction": "out",
    "legend.fontsize": 14,
    "grid.color": ".8",
    "grid.linestyle": "-",
    "grid.linewidth": 1,
}

rx_color = "red"
bg_color = "blue"
dc_color = "darkgoldenrod"


# set default styles for plotting
set_defaults()



"""Plotting class for sequence alignments."""

from rnavigate import plots


class Alignment(plots.Plot):
    """Class for plotting sequence alignments.

    Parameters
    ----------
    num_samples : int
        The number of samples to plot. Always 2.
    rows : int, optional
        The number of rows of plots. Always 1.
    cols : int, optional
        The number of columns of plots. Always 1.
    **kwargs : dict
        Additional keyword arguments to pass to plots.Plot.

    Attributes
    ----------
    region : tuple
        The region of the alignment to plot. Defaults to (1, len(alignment1))
    fig : matplotlib.figure.Figure
        The figure containing the plot.
    axes : numpy.array
        A 1x1 array of the axes containing the plot.
    """

    def __init__(self, num_samples, rows=None, cols=1, **kwargs):
        """Create a new Alignment object."""
        super().__init__(num_samples, rows, cols, **kwargs)

    def plot_data(self, alignment, label, ax=None):
        """Add the alignment to the next (or given) axes.

        Parameters
        ----------
        alignment : rnavigate.Alignment
            The alignment to plot.
        label : str
            The label for the alignment.
        ax : matplotlib.axes.Axes, optional
            The axes containing the plot.
            Defaults to None.
        """
        if ax is None:
            ax = self.get_ax()
        alignment1 = alignment.alignment1
        self.region = (1, len(alignment1))
        ax.set(xlim=(0, len(alignment1) + 1), ylim=(-5, 5), yticks=[])
        plots.plot_sequence_alignment(ax, alignment, label)
        for spine in ["top", "bottom", "left", "right"]:
            ax.spines[spine].set_color(None)

    def set_figure_size(
        self,
        fig=None,
        ax=None,
        rows=None,
        cols=None,
        height_ax_rel=0.03,
        width_ax_rel=0.03,
        width_ax_in=None,
        height_ax_in=None,
        height_gap_in=1,
        width_gap_in=0.5,
        top_in=1,
        bottom_in=0.5,
        left_in=0.5,
        right_in=0.5,
    ):
        """Set the figure size for the plot.

        Parameters
        ----------
        fig : matplotlib.figure.Figure, optional
            The figure containing the plot.
            Defaults to None.
        ax : matplotlib.axes.Axes, optional
            The axes containing the plot.
            Defaults to None.
        rows : int, optional
            The number of rows of plots.
            Defaults to None.
        cols : int, optional
            The number of columns of plots.
            Defaults to None.
        height_ax_rel : float, optional
            The relative height of each axes.
            Defaults to 0.03.
        width_ax_rel : float, optional
            The relative width of each axes.
            Defaults to 0.03.
        width_ax_in : float, optional
            The width of each axes in inches.
            Defaults to None.
        height_ax_in : float, optional
            The height of each axes in inches.
            Defaults to None.
        height_gap_in : float, optional
            The height of the gap between axes in inches.
            Defaults to 1.
        width_gap_in : float, optional
            The width of the gap between axes in inches.
            Defaults to 0.5.
        top_in : float, optional
            The top margin in inches.
            Defaults to 1.
        bottom_in : float, optional
            The bottom margin in inches.
            Defaults to 0.5.
        left_in : float, optional
            The left margin in inches.
            Defaults to 0.5.
        right_in : float, optional
            The right margin in inches.
            Defaults to 0.5.
        """
        super().set_figure_size(
            fig=fig,
            ax=ax,
            rows=rows,
            cols=cols,
            height_ax_rel=height_ax_rel,
            width_ax_rel=width_ax_rel,
            width_ax_in=width_ax_in,
            height_ax_in=height_ax_in,
            height_gap_in=height_gap_in,
            width_gap_in=width_gap_in,
            top_in=top_in,
            bottom_in=bottom_in,
            left_in=left_in,
            right_in=right_in,
        )



#!/usr/bin/env python3
# -----------------------------------------------------
# fpocket analysis code
# Seth Veenbaas, Patrick Irving
# Weeks Lab, UNC-CH
# 2-9-2023
#
# Version 1.0.0
#
# -----------------------------------------------------
import argparse
import os
import glob
import pickle
import pandas as pd
from pymol import cmd
from prody import *
from fpocketR import analyze, pocket, figures, util
confProDy(verbosity='none')
# -----------------------------------------------------


def pipeline(
    pdb : str,
    ss : str,
    chain : str,
    state : int,
    ligand : str,
    ligandchain : str,
    knownnt : list[int],
    offset : int,
    qualityfilter : float,
    m : float,
    M : float,
    i : int,
    D : float,
    A : float,
    p : float,
    out : str,
    name : str,
    dpi : int,
    yes : bool,
    zoom : float,
    connectpocket : bool,
    alignligand : str,
):   
    """Runs pocket finding pipeline

    Args:
        pdb (str): Path to input .pdb file.
        ss (str)): Path to input secondary structure drawing.
        chain (str): Chain identifier for desired RNA chain (default='A').
        state (int): Structural state to analyze.
        ligandchain (str): Chain identifier for desired ligand (default=chain).
        offset (int): Sequence offset between .pdb and .ss file (default=None).
        qualityfilter (float): Minimum fpocket score filter for pockets.
        m (float): Min. a-sphere radius in angstroms (default=3.0).
        M (float): Max. a-sphere radius in angstroms (default=5.7).
        i (int): Min. number of a-spheres per pocket (default=42).
        D (float): a-sphere clustering distance in angstroms (default=1.65).
        A (int): # of electroneg. atoms to define a polar a-sphere (default=3).
        p (float): Max. ratio of apolar a-spheres in a pocket (default=0.0).
        out (str): name of fpocket output parent directory name.
        name (str): Output file name prefix (default={pdb_name}).
        dpi (int): Figure resolution in dpi (default=300).
        ligand (str): Ligand residue name (usually a 3-letter code).
        yes (boolean): Overwrite output files and directories with same name.
        zoom (float): Zoom buffer distance (Å) for creating 3D figures.
        connectpocket (boolean): Connects pockets in 2D figure (Default=False).
        alignligand (str): Align ligand to pymol output (Default=True).

    Returns:
        str: Path to clean .pdb input file.
        str: Path directory contianing fpocket outputs for analysis.
        object: Pandas Dataframe with characteristics for each pocket.

    """

    # Checks if required input files are accessible/exist.
    print('Checking input files.')
    util.is_accessible(pdb, 'pdb')

    if chain is None:
        chain = util.get_first_rna_chain(pdb)
    else:
        util.is_rna_chain(pdb, chain)

    # Runs fpocket on input pdb file and manages output files.
    analysis, yes = pocket.find_pockets(
        pdb,
        chain,
        state,
        m,
        M,
        i,
        D,
        A,
        p,
        out,
        name,
        yes,
    )

    # Checks if the analysis directory is accessible.
    util.is_accessible(analysis, 'analysis directory')

    # Get paths to fpocket input and output file.
    (
        pdb,
        pdb_out,
        pqr_out,
        info_txt,
        pockets_out,
        pdb_code,
        name,
        ) = util.get_file_paths(analysis, name, pdb, state)

    # Analyze fpocket data and create pocket characteristics dataframe.
    (pc_df, rna_coords) = analyze.analyze_pockets(
        pdb,
        pqr_out,
        pdb_out,
        analysis,
        name,
        info_txt,
        pockets_out,
        pdb_code,
        chain,
        state,
        ligandchain,
        ligand,
        m,
        M,
        i,
        D,
        A,
        p,
        qualityfilter,
        knownnt,
    )
    
    offset = util.get_offset(pdb, chain, offset) if offset is None else offset

    # Generates 1D (.csv), 2D (.png, .svg), and 3D (.pdb, .pse, .png)
    pocket_cmap = figures.make_figures(
        pdb,
        state,
        pc_df,
        rna_coords,
        ss,
        analysis,
        name,
        chain,
        dpi,
        zoom,
        offset,
        connectpocket,
        alignligand
    )

    return pc_df, out, pocket_cmap, chain, yes


# -----------------------------------------------------
def parseArgs():
    prs = argparse.ArgumentParser()
    pocket_type = prs.add_mutually_exclusive_group()

# Input options
    prs.add_argument(
        '-pdb',
        '--pdb',
        type=str,
        required=True,
        default=None,
        help='Path to a .pdb file, .cif file, or 4 charater PDB '
        'indentification code.',
    )
    prs.add_argument(
        '-ss',
        '--ss',
        type=str,
        required=False,
        default=None,
        help='Path to an .ss or other secondary structure file '
        'for generating secondary structure figures.',
    )
    
# fpocket parameter options
    prs.add_argument(
        '-m',
        type=float,
        required=False,
        default=3.00,
        help='fpocket -m flag. Specifies the minimum radius '
        'for an a-sphere (3.0).',
    )
    prs.add_argument(
        '-M',
        '--M',
        type=float,
        required=False,
        default=5.70,
        help='fpocket -M flag. Specifies the maximium radius '
        'for an a-sphere (5.70).',
    )
    prs.add_argument(
        '-i',
        '--i',
        type=int,
        required=False,
        default=42,
        help='fpocket -i flag. Specifies the minimum number '
        'of a-spheres per pocket (42).',
    )
    prs.add_argument(
        '-D',
        '--D',
        type=float,
        required=False,
        default=1.65,
        help='fpocket -D flag. Specifies the a-sphere '
        'clustering distance for forming pockets (1.65).'
    )
    prs.add_argument(
        '-A',
        '--A',
        type=int,
        required=False,
        default=3,
        help='fpocket -A flag. Number of electronegative atoms '
        'required to define a polar a-sphere (3).',
    )
    prs.add_argument(
        '-p',
        '--p',
        type=float,
        required=False,
        default=0.0,
        help='fpocket -p flag. Maximum ratio of apolar a-spheres '
        'in a pocket (0.0).',
    )
    
# Output options
    prs.add_argument(
        '-o',
        '--out',
        type=str,
        required=False,
        default=None,
        help='Path to the output parent directory \
            (Default: "./fpocketR_out").',
    )
    prs.add_argument(
        '-n',
        '--name',
        type=str,
        required=False,
        default=None,
        help='Specify output filename prefix and output subdirectory name \
            (Default: "{PDB}_clean_out").',
    )
    prs.add_argument(
        '-y',
        '--yes',
        required=False,
        action='store_true',
        help='Answers yes to user prompts for overwriting files (False).',
    )

# Analysis options
    prs.add_argument(
        '-s',
        '--state',
        type=int,
        required=False,
        default=None,
        help='Specify the NMR states/model you would like to analyze. 0 for all (None).',
    )
    prs.add_argument(
        '-c',
        '--chain',
        type=str,
        required=False,
        default=None,
        help='Specify a chain from the input .pdb file \
            (Default: <first_rna_chain>).',
    )
    pocket_type.add_argument(
        '-l',
        '--ligand',
        type=str,
        required=False,
        default=None,
        help='PDB ligand identification code (2-3 characters).',
    )
    prs.add_argument(
        '-lc',
        '--ligandchain',
        type=str,
        required=False,
        default=None,
        help='Chain containing ligand the from the input .pdb file (--chain input).',
    )
    pocket_type.add_argument(
        '-nt',
        '--knownnt',
        type=parse_int,
        required=False,
        default=None,
        help='List residue ID of nts in known pocket (e.g. 1,2,3) (None).',
    )
    prs.add_argument(
        '-off',
        '--offset',
        type=int,
        required=False,
        default=None,
        help='Specify an offset between the '
        'starting nucleotide of the rna sequence and '
        'starting nucleotide of the PDB structure.\n'
        'offset = starting index of the PDB sequence - 1\n'
        '(automatic)'
    )
    prs.add_argument(
        '-qf',
        '--qualityfilter',
        type=float,
        required=False,
        default=0.0,
        help='Specify minimum fpocket score for pocket (0.0).',
    )
    
# Figure options
    prs.add_argument(
        '-dpi',
        '--dpi',
        type=int,
        required=False,
        default=300,
        help='Sets figure resolution in dpi (300).',
    )
    prs.add_argument(
        '-z',
        '--zoom',
        type=float,
        required=False,
        default=5.0,
        help='Zoom buffer (Å) for creating 3D figures (5.0).',
    )
    prs.add_argument(
        '-cp',
        '--connectpocket',
        required=False,
        action='store_true',
        help='Visually connects pockets in 2D figures (False).',
    )
    prs.add_argument(
        '-al',
        '--alignligand',
        type=str,
        required=False,
        help='Align structure with pocket prediction (target structure) \
            to an RNA structure with a ligand (mobile structure). \
            If `str`: path to a .pdb file, .cif file, or 4 character \
            PDB identification of the mobile structure. \
            If `bool`: input PDB file is used as the mobile structure.',
    )

    args = prs.parse_args()
    return args

def parse_int(string : str) -> list[int]:
    try:
        # Split the string by commas and convert each part to an integer
        return [int(item) for item in string.split(',')]
    except ValueError:
        raise argparse.ArgumentTypeError("List of integers expected. Example: '1,2,3,4'")


def main(
    pdb : str,
    ss : str,
    chain : str,
    state : int,
    ligand : str,
    ligandchain : str,
    knownnt : list[int],
    offset : int,
    qualityfilter : float,
    m : float,
    M : float,
    i : int,
    D : float,
    A : float,
    p : float,
    out : str,
    name : str,
    dpi : int,
    yes : bool,
    zoom : float,
    connectpocket : bool,
    alignligand : str,
):
    """Runs the fpocket analysis pipeline.
    Pipeline runs once: by default or if provided a user specified state
    is specified using the -s flag.

    Pipeline runs multiple times: if the -s flag is set to 0 (all).
    This feature is intended for analyzing NMR structures
    with several modeled states.
    """

    # Check if pdb contains a file extension.
    if len(pdb.split('.')) < 2:
        pdb = util.fetch_pdb(pdb)

    # Set pdb name
    if name is None:
        name = os.path.basename(pdb)[0:-4]

    # Set structure to align to.
    if alignligand == 'False' or alignligand == 'none':
        alignligand = None
    elif alignligand is None or alignligand == 'True':
        alignligand = pdb
    elif len(alignligand.split('.')) < 2:
        alignligand = util.fetch_pdb(alignligand)
    elif not os.path.isfile(alignligand):
        alignligand = None

    # Runs pipeline for a single state of the input structure.
    if state != 0:
        if out is None:
            out = f'fpocketR_out'
        (_, _, _, _, _) = pipeline(
            pdb,
            ss,
            chain,
            state,
            ligand,
            ligandchain,
            knownnt,
            offset,
            qualityfilter,
            m,
            M,
            i,
            D,
            A,
            p,
            out,
            name,
            dpi,
            yes,
            zoom,
            connectpocket,
            alignligand,
        )

    # Runs pipeline for multiple states of the input structure.
    elif state == 0:
        if out is None:
            out = f'Multistate_{pdb.split(".")[0]}'

        try:
            structure = parsePDB(pdb)
            num_states = structure.numCoordsets()
        except:
            print('ERROR: Unable to perform multisate analysis.\n'
                  f'The header for {pdb} does not contain state information.\n')
            exit() 

        state_tracker_filename = f"{out}/state_tracker.txt"
        last_state = util.get_last_processed_state(state_tracker_filename)  # Get the last processed state
        if last_state > 0:
            start_state = last_state
        else:
            start_state = 1
        for state in range(start_state, num_states + 1):
            print(f'\nFinding pockets in state {state}/{num_states}...\n')
            (pc_df, out, pocket_cmap, chain, yes) = pipeline(
                pdb,
                ss,
                chain,
                state,
                ligand,
                ligandchain,
                knownnt,
                offset,
                qualityfilter,
                m,
                M,
                i,
                D,
                A,
                p,
                out,
                name,
                dpi,
                yes,
                zoom,
                connectpocket,
                alignligand,
            )
            yes = yes
            util.update_last_processed_state(state_tracker_filename, state)
            # pc_all_states = pd.concat([pc_all_states, pc_df])
            # multistate_pocket_cmap[state]=pocket_cmap
                

        # Generates csv output containing pocket characteristics for all states.
        pc_files = glob.glob(f"{out}/*/*_out_pocket_characteristics.csv")
        pc_files = util.natsorted(pc_files)

        # Check if any files match
        if not pc_files:
            raise FileNotFoundError(f"No files matching '*_out_pocket_characteristics.csv' found in {out}")
        # Read and concatenate all matching CSV files into a single DataFrame
        pc_all_states = pd.concat((pd.read_csv(file) for file in pc_files), ignore_index=True)
        pc_all_states.to_csv(
            f'{out}/{name}_all_states_pocket_characteristics.csv',
            index=False, float_format='%.2g')
        
        pocket_cmap_files = glob.glob(f"{out}/*/*_maps.pkl")
        multistate_pocket_cmap = {}
        multistate_pocket_nt_color = {}
        # Loop through all files and load the dictionaries
        for file_path in pocket_cmap_files:
            with open(file_path, 'rb') as file:
                # Load the JSON data (assumed to be a dictionary)
                data = pickle.load(file)
                # Merge the current dictionary into the combined dictionary
                multistate_pocket_cmap.update(data[0])
                multistate_pocket_nt_color.update(data[1])
        # Sort the combined dictionary by keys (ascending order)
        multistate_pocket_cmap = dict(sorted(multistate_pocket_cmap.items()))
        
        if ss:
            # Generates a 2D for pockets in all states.
            print(f'Making all states 2D figure...')
            figures.get_all_states_2D_figure(
                name,
                out,
                ss,
                num_states,
                multistate_pocket_nt_color,
            )

        # Generates a 3D for pockets in all states.
        print(f'Making all states 3D figure...')
        
        figures.get_all_states_3D_figure(
            num_states,
            out,
            name,
            multistate_pocket_cmap,
            alignligand,
            dpi,
            chain,
            zoom
        )

    # Close pymol session.
    cmd.quit()


if __name__ == "__main__":
    main(**vars(parseArgs()))



"""Classes for storing and manipulating data for RNAvigate.

This module contains the base classes for RNAvigate data classes:
    Sequence: represents a nucleotide sequence
    Data: represents a data table with a sequence
"""

from os.path import isfile
import Bio.SeqIO
import numpy as np
import pandas as pd
import matplotlib.colors as mpc
from rnavigate import styles
from rnavigate import data


def normalize_sequence(sequence, t_or_u="U", uppercase=True):
    """Returns sequence as all uppercase nucleotides and/or corrects T or U.

    Parameters
    ----------
    sequence : string or RNAvigate Sequence)
        The sequence
        If given an RNAvigate Sequence, the sequence string is retrieved
    t_or_u : "T", "U", or False, defaults to "U"
        "T" converts "U"s to "T"s
        "U" converts "T"s to "U"s
        False does nothing
    uppercase bool, defaults to True
        Whether to make sequence all uppercase

    Returns
    -------
        string
            the cleaned-up sequence string
    """
    if isinstance(sequence, Sequence):
        sequence = sequence.sequence
    if uppercase:
        sequence = sequence.upper()
    if not t_or_u:
        pass
    elif t_or_u.upper() == "U":
        sequence = sequence.replace("t", "u").replace("T", "U")
    elif t_or_u.upper() == "T":
        sequence = sequence.replace("u", "t").replace("U", "T")
    return sequence


class Sequence:
    """A class for storing and manipulating RNA sequences.

    Parameters
    ----------
    sequence : string or pandas.DataFrame
        sequence string, fasta file, or a Pandas dataframe containing a
        "Sequence" column
    name : string, optional
        The name of the sequence, defaults to None

    Attributes
    ----------
    sequence : string
        The sequence string
    name : string
        The name of the sequence
    other_info : dict
        A dictionary of additional information about the sequence
    null_alignment : SequenceAlignment
        An alignment of the sequence to itself
    """

    def __init__(self, input_data, name=None):
        """Initialize the Sequence object."""
        self.name = name
        self.other_info = dict()
        if isinstance(input_data, str):
            if isfile(input_data):
                self.sequence = self.read_fasta(input_data)
            else:
                self.sequence = input_data
        elif isinstance(input_data, pd.DataFrame):
            self.sequence = self.get_seq_from_dataframe(input_data)
        elif isinstance(input_data, Sequence):
            self.sequence = input_data.sequence
        self.null_alignment = data.SequenceAlignment(self, self)

    def __str__(self):
        """Return the name of the sequence."""
        if self.name is None:
            return "seq-object"
        return self.name

    def read_fasta(self, fasta):
        """Parse a fasta file for the first sequence.

        Parameters
        ----------
        fasta : string
            path to fasta file
        """
        with open(fasta, "r") as file:
            fasta = list(Bio.SeqIO.parse(file, "fasta"))
        return str(fasta[0].seq).replace("T", "U")

    def get_seq_from_dataframe(self, dataframe):
        """Parse a dataframe for the sequence string, store as self.sequence.

        Parameters
        ----------
        dataframe : pandas.DataFrame
            must contain a "Sequence" column
        """
        sequence = "".join(dataframe["Sequence"].values)
        return sequence.replace("T", "U").replace("t", "u")

    @property
    def length(self):
        """Get the length of the sequence

        Returns
        -------
        length : int
            the length of self.sequence
        """
        return len(self.sequence)

    def normalize_sequence(self, t_or_u="U", uppercase=True):
        """Converts sequence to all uppercase nucleotides and corrects T or U.

        Parameters
        ----------
        t_or_u : "T", "U", or False, defaults to "U"
            "T" converts "U"s to "T"s
            "U" converts "T"s to "U"s
            False does nothing.
        uppercase : bool, defaults to True
            Whether to make sequence all uppercase
        """
        self.sequence = normalize_sequence(self, t_or_u=t_or_u, uppercase=uppercase)

    def get_aligned_data(self, alignment):
        """Get a copy of the sequence positionally aligned to another sequence.

        Parameters
        ----------
        alignment : rnavigate.data.Alignment
            the alignment to use

        Returns
        -------
        aligned_sequence : rnavigate.data.Sequence
            the aligned sequence
        """
        return Sequence(alignment.target_sequence)

    def get_colors_from_sequence(self):
        """Get a colors and colormap representing the nucleotide sequence.

        Returns
        -------
        colors : numpy array
            one matplotlib color-like value for each nucleotide in self.sequence
        colormap : rnavigate.data.ScalarMappable
            a colormap used for creating a colorbar
        """
        colors = np.array([styles.get_nt_color(nt) for nt in self.sequence])
        colormap = data.ScalarMappable(
            cmap=[styles.get_nt_color(nt) for nt in "AUGC"],
            normalization="none",
            values=None,
            extend="neither",
            title="Nucleotide identity",
            alpha=1,
            ticks=[0, 1, 2, 3],
            tick_labels=["A", "U", "G", "C"],
        )
        return colors, colormap

    def get_colors_from_positions(self, pos_cmap="rainbow"):
        """Get colors and colormap representing the nucleotide position.

        Parameters
        ----------
        pos_cmap : str, defaults to "rainbow"
            cmap used for position colors

        Returns
        -------
        colors : numpy array
            one matplotlib color-like value for each nucleotide in self.sequence
        colormap : rnavigate.data.ScalarMappable
            a colormap used for creating a colorbar
        """
        colormap = data.ScalarMappable(
            cmap=pos_cmap,
            normalization="min_max",
            values=[1, self.length],
            extend="neither",
            title="Nucleotide position",
            alpha=1,
        )
        colors = colormap.values_to_hexcolors(np.arange(self.length))
        return colors, colormap

    def get_colors_from_profile(self, profile):
        """Get colors and colormap representing per-nucleotide data.

        Parameters
        ----------
        profile : rnavigate.data.Profile
            the profile to use to get colors.

        Returns
        -------
        colors : numpy array
            one matplotlib color-like value for each nucleotide in self.sequence
        colormap : rnavigate.data.ScalarMappable
            a colormap used for creating a colorbar
        """
        alignment = data.SequenceAlignment(profile, self)
        colors = alignment.map_values(profile.colors, fill="#808080")
        colormap = profile.cmap
        return colors, colormap

    def get_colors_from_annotations(self, annotations, default_color="gray"):
        """Get colors and colormap representing sequence annotations.

        Parameters
        ----------
        annotations : list of rnavigate.data.Annotations
            the annotations to use to get colors.
        default_color : matplotlib color-like, defaults to "gray"
            the color to use for nucleotides not in any annotation

        Returns
        -------
        colors : numpy array
            one matplotlib color-like value for each nucleotide in self.sequence
        colormap : rnavigate.data.ScalarMappable
            a colormap used for creating a colorbar
        """
        colors = np.full(self.length, default_color, dtype="<U16")
        cmap = [default_color]
        tick_labels = ["other"]
        for annotation in annotations:
            cmap.append(annotation.color)
            tick_labels.append(annotation.name)
            annotation = annotation.get_aligned_data(
                data.SequenceAlignment(annotation, self)
            )
            for site in annotation.get_sites():
                colors[site - 1] = annotation.color
        colormap = data.ScalarMappable(
            cmap=cmap,
            normalization="none",
            values=None,
            extend="neither",
            title="Annotations",
            alpha=1,
            ticks=list(range(len(annotations) + 1)),
            tick_labels=tick_labels,
        )
        return colors, colormap

    def get_colors_from_structure(self, structure):
        """Get colors and colormap representing base-pairing status.

        Parameters
        ----------
        structure : rnavigate.data.SecondaryStructure
            the structure to use to get colors.

        Returns
        -------
        colors : numpy array
            one matplotlib color-like value for each nucleotide in self.sequence
        colormap : rnavigate.data.ScalarMappable
            a colormap used for creating a colorbar
        """
        cmap = ["darkOrange", "darkOrchid", "gray"]
        ct_colors = [cmap[int(pair == 0)] for pair in structure.pair_nts]
        alignment = data.SequenceAlignment(structure, self)
        colors = alignment.map_values(ct_colors, fill="gray")
        colormap = data.ScalarMappable(
            cmap=cmap,
            normalization="none",
            values=None,
            extend="neither",
            title="Base-pairing status",
            alpha=1,
            ticks=[0, 1, 2],
            tick_labels=["unpaired", "paired", "unaligned"],
        )
        return colors, colormap

    def get_colors(
        self, source, pos_cmap="rainbow", profile=None, structure=None, annotations=None
    ):
        """Get colors and colormap representing information about the sequence.

        Parameters
        ----------
        source : str, list, or matplotlib color-like
            the source of the color information
            if a string, must be one of:
                "sequence", "position", "profile", "structure", "annotations"
            if a list, must be a list of matplotlib color-like values, colormap
                will be None.
            if a matplotlib color-like value, all nucleotides will be colored
                that color, colormap will be None.
        pos_cmap : str, defaults to "rainbow"
            cmap used for position colors if source is "position"
        profile : rnavigate.data.Profile, optional
            the profile to use to get colors if source is "profile"
        structure : rnavigate.data.SecondaryStructure, optional
            the structure to use to get colors if source is "structure"
        annotations : list of rnavigate.data.Annotations, optional
            the annotations to use to get colors if source is "annotations"

        Returns
        -------
        colors : numpy array
            one matplotlib color-like value for each nucleotide in self.sequence
        colormap : rnavigate.data.ScalarMappable
            a colormap used for creating a colorbar
        """
        if isinstance(source, str) and (source == "sequence"):
            return self.get_colors_from_sequence()
        elif isinstance(source, str) and (source == "position"):
            return self.get_colors_from_positions(pos_cmap=pos_cmap)
        elif isinstance(source, str) and (source == "profile"):
            return self.get_colors_from_profile(profile=profile)
        elif isinstance(source, str) and (source == "annotations"):
            return self.get_colors_from_annotations(annotations=annotations)
        elif isinstance(source, str) and (source == "structure"):
            return self.get_colors_from_structure(structure)
        elif mpc.is_color_like(source):
            return np.full(self.length, source, dtype="<U16"), None
        elif (len(source) == self.length) and all(mpc.is_color_like(c) for c in source):
            return np.array(list(source)), None
        else:
            print("Invalid colors:")
            print("\tchoices are profile, sequence, position, structure,")
            print("annotations, a list of mpl colors, or a single mpl color.")
            print("Defaulting to sequence.")
            return self.get_colors_from_sequence()


class Data(Sequence):
    """The base class for RNAvigate Profile and Interactions classes.

    Parameters
    ----------
    input_data : pandas.DataFrame or str
        a pandas dataframe or path to a data file
    sequence : string or rnavigate.data.Sequence
        the sequence to use for the data
    metric : string or dict
        the column of the dataframe to use as the default metric to visualize
    metric_defaults : dict
        a dictionary of metric defaults
    read_table_kw : dict, optional
        kwargs dictionary passed to pd.read_table
    name : string, optional
        the name of the data, defaults to None

    Attributes
    ----------
    data : pandas.DataFrame
        the data table
    filepath : string
        the path to the data file
    sequence : string or rnavigate.data.Sequence
        the sequence to use for the data
    metric : string or dict
        the column of the dataframe to use as the metric to visualize
    metric_defaults : dict
        A dictionary of metric values and default settings for visualization
    default_metric : string
        the default metric to use for visualization
    """

    def __init__(
        self,
        input_data,
        sequence,
        metric,
        metric_defaults,
        read_table_kw=None,
        name=None,
    ):
        """Initialize the Data object."""
        if read_table_kw is None:
            read_table_kw = {}
        # assign data
        if isinstance(input_data, pd.DataFrame):
            self.data = input_data
            self.filepath = "dataframe"
        elif isfile(input_data):
            self.data = self.read_file(input_data, read_table_kw)
            self.filepath = input_data
        else:
            print(f"{self} initialized without data.")
        # assign sequence
        if sequence is None:
            sequence = self.data
        super().__init__(sequence, name=name)
        # assign metrics
        self.metric_defaults = {
            "default": {
                "metric_column": "Profile",
                "error_column": None,
                "color_column": None,
                "cmap": "viridis",
                "normalization": "0_1",
                "values": None,
                "extend": "neither",
                "title": "Type: metric",
                "alpha": 0.7,
            },
            "Distance": {
                "metric_column": "Distance",
                "error_column": None,
                "color_column": None,
                "cmap": "cool",
                "normalization": "min_max",
                "values": [5, 50],
                "extend": "both",
                "title": "3D distance",
                "alpha": 0.7,
            },
        }
        self.add_metric_defaults(metric_defaults)
        self.default_metric = metric
        self._metric = None
        self.metric = metric

    def add_metric_defaults(self, metric_defaults):
        """Add metric defaults to self.metric_defaults"""
        default_defaults = self.metric_defaults["default"]
        for metric, defaults in metric_defaults.items():
            self.metric_defaults[metric] = default_defaults | defaults

    @property
    def metric(self):
        """Get the column of the dataframe to use as the metric for visualization."""
        return self._metric["metric_column"]

    @metric.setter
    def metric(self, value):
        """Set the column of the dataframe and other settings to use for visualization.

        Parameters
        ----------
        value : string or dict
            the column of the dataframe to use as the metric for visualization
            if a string, must be one of:
                "default", "Distance", or a column name in self.data
            if a dict, must contain at least one of:
                "metric_column", "error_column", "color_column", "cmap",
                "normalization", "values", "extend", "title", or "alpha"
        """
        if isinstance(value, str):
            try:
                self._metric = self.metric_defaults[value]
                return
            except KeyError as exception:
                if value not in self.data.columns:
                    print(
                        f"metric ({value}) not found in data:\n"
                        + str(list(self.data.columns))
                    )
                    raise exception
                self._metric = self.metric_defaults["default"]
                self._metric["metric_column"] = value
                return
        try:
            defaults = self.metric_defaults[value["metric_column"]]
        except KeyError:
            defaults = self.metric_defaults["default"]
        for key in value:
            if key not in defaults:
                raise ValueError(
                    f"{key} is not an expected value for metric "
                    f"setting:\n{list(defaults.keys())}"
                )
        self._metric = defaults | value

    @property
    def error_column(self):
        """Get the column of the dataframe to use as the error for visualization."""
        if self._metric["error_column"] in self.data.columns:
            return self._metric["error_column"]
        print(f"Warning: {self} missing expected error column")
        return None

    @property
    def color_column(self):
        """Get the column of the dataframe to use as the color for visualization."""
        if self._metric["color_column"] in self.data.columns:
            return self._metric["color_column"]
        return self._metric["metric_column"]

    @property
    def cmap(self):
        """Get the colormap to use for colorbars and to retrieve colors."""
        cmap_kwargs = {}
        for k, v in self._metric.items():
            if not k.endswith("column"):
                cmap_kwargs[k] = v
        return data.ScalarMappable(**cmap_kwargs)

    @property
    def colors(self):
        """Get one matplotlib color-like value for each nucleotide in self.sequence."""
        values = self.data[self.color_column]
        if values.dtype.name == "bool":
            values = values.astype("int")
        return self.cmap.values_to_hexcolors(np.ma.masked_invalid(values))

    def read_file(self, filepath, read_table_kw):
        """Convert data file to pandas dataframe and store as self.data

        Parameters
        ----------
        filepath : string
            path to data file containing interactions
        read_table_kw : dict
            kwargs dictionary passed to pd.read_table

        Returns
        -------
        dataframe : pandas.DataFrame
            the data table
        """
        return pd.read_table(filepath, **read_table_kw)



#!/usr/bin/env python3
# -----------------------------------------------------------------------------
# Module for controlling fpocket
# Seth Veenbaas
# Weeks Lab, UNC-CH
# 2022
#
# Version 1.3.0
#
# -----------------------------------------------------------------------------

import os
import sys
import subprocess
import shutil
import time
from pymol import cmd


def find_pockets(
    pdb : str,
    chain : str,
    state : int,
    m : float,
    M : float,
    i :int,
    D : float,
    A : int,
    p : float,
    out : str,
    name : str,
    yes : bool,
    ) -> str:
    """Pocket finding pipeline:
        - cleans pdb file to generate rna-only file
        - runs pocket prediction using fpocket
        - manages fpocket output files

    Args:
        pdb (str): Path to input .pdb file.
        chain (str): Chain identifier for desired RNA chain (default='A').
        state (int): Structural state to analyze.
        m (float): Min. a-sphere radius in angstroms (default=3.0).
        M (float): Max. a-sphere radius in angstroms (default=5.7).
        i (int): Min. number of a-spheres per pocket (default=42).
        D (float): a-sphere clustering distance in angstroms (default=1.65).
        out (str): name of fpocket output parent directory name.
        yes (boolean): Overwrite output files and directories with same name.

    Returns:
        str: path to output directory contianing fpocket outputs for analysis
    """

    # Path to clean (ligand-free) pdb file
    pdb_clean = f'{name}_clean.pdb'

    # Makes pdb_clean if it is not already a file
    if not os.path.isfile(pdb_clean):
        clean_pdb(pdb, pdb_clean)

    # Runs fpocket on cleaned pdb file.
    run_fpocket(pdb_clean, name, chain, state, m, M, i, D, A, p)

    # Files fpocket outputs into directories and manages overwriting.
    analysis, yes = file_fpocket(pdb_clean, state, out, yes)
    return analysis.strip('/'), yes

# -----------------------------------------------------------------------------


def clean_pdb(pdb : str, pdb_clean : str) -> None:
    """Cleans a .pdb file input and saves output as a .pdb file.
       Removes not polymer molecules (ligands) and proteins.
       Preserves modified/heteroatom RNA residues.

    Args:
        pdb (str): path to input .pdb file.
        pdb_clean (str): path to output (cleaned) .pdb file.
    """
    cmd.load(pdb)
    cmd.alter('polymer', 'type="ATOM"')
    cmd.save(pdb, state='0')
    time.sleep(1)
    cmd.remove('not polymer')
    cmd.remove('byres polymer & name CA')
    cmd.save(pdb_clean, state='0')
    time.sleep(1)
    cmd.reinitialize()


# class MissingEnvironmentVariable(Exception):
#     pass

def run_fpocket(
    pdb : str,
    name : str,
    chain :str,
    state : int,
    m : float,
    M : float,
    i :int,
    D : float,
    A : int,
    p : float,
    ) -> None:
    """Detects potential binding pockets in RNA structures using fpocket.

    Args:
        pdb (str): Path to input .pdb file.
        chain (str): Chain identifier for desired RNA chain (default='A').
        state (int): Structural state to analyze.
        m (float): Min. a-sphere radius in angstroms (default=3.0).
        M (float): Max. a-sphere radius in angstroms (default=5.7).
        i (int): Min. number of a-spheres per pocket (default=42).
        D (float): a-sphere clustering distance in angstroms (default=1.65).
    """
    # Prints announcement that fpocket is searching for pockets.
    name = os.path.basename(pdb)[0:-4]
    print(f'***** POCKET HUNTING {name} *****')
    # Runs fpocket bash commands
    # Try to find fpocket in the current conda/mamba/micromamba environment
    env_prefix = os.environ.get('CONDA_PREFIX', os.environ.get('MAMBA_ROOT_PREFIX', sys.prefix))
    fpocket_env_path = os.path.join(env_prefix, 'bin', 'fpocket')
    if os.path.isfile(fpocket_env_path) and os.access(fpocket_env_path, os.X_OK):
        fpocket_path = fpocket_env_path
    else:
        fpocket_path = shutil.which('fpocket')
    if not fpocket_path:
        raise FileNotFoundError('fpocket executable not found in current environment or PATH. Ensure fpocket is installed and available.')
    cmd = [fpocket_path, '-f', pdb, '-k', chain, '-l', str(state), '-m', str(m), '-M', str(M), '-i', str(i), '-D', str(D), '-A', str(A), '-p', str(p), '-w', 'p']
    process = subprocess.Popen(cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE, text=True)
    stdout, stderr = process.communicate()
    print(stdout)
    print(stderr)
    if 'EnvironmentLocationNotFound' in stdout or 'EnvironmentLocationNotFound' in stderr:
        raise OSError('Unable to run fpocket because the fpocketR environment does not exist.\nInstall the fpocketR environment.')


def file_fpocket(pdb : str, state : int, out : str, yes : bool) -> tuple[str, bool]:
    """Moves fpocket outputs into designated output directory.
       Default directory name specifies the fpocket parameters used.
       Manages overwriting files/directories if thet already exist.

    Args:
        pdb (str): Path to input .pdb file.
        state (int): Structural state to analyze.
        out (str): name of fpocket output parent directory name.
        y (boolean): Overwrites output files and directories with same name.

    Returns:
        str: path to output directory contianing fpocket outputs for analysis
    """
    # Moves fpocket output directories into a shared directory.
    source_dir = f'{pdb.rsplit(".")[0]}_out'
        
    if not os.path.isdir(source_dir):
        raise FileNotFoundError(f'fpocket output directory does not exist: {source_dir}.')

    if state is None:
        dest_dir = os.path.join(out, f'{pdb.rsplit(".")[0]}_out')
    else:
        dest_dir = os.path.join(out, f'{pdb.rsplit(".")[0]}_state{state}_out')

    if source_dir == dest_dir:
        analysis = source_dir

    elif not os.path.isdir(dest_dir):
        analysis = shutil.move(source_dir, dest_dir)

    # Prompts user to overwrite an existing file with same name.
    elif os.path.isdir(dest_dir):
        if not yes:
            remove = input(
                'A directory already exists with this name.\n'
                f'{dest_dir}\n\n'
                'Overwrite directory? [y/n]: '
            )
            print()
        if yes or remove in ('y', 'Y', 'yes', 'Yes'):
            yes = True
            shutil.rmtree(dest_dir)
            analysis = shutil.move(source_dir, dest_dir)
        else:
            print(
                'Exiting program. \n'
                'The name of the output directory can '
                'be changed with the --name flag.'
            )
            exit()

    # Adds state identifier number to file names.
    if state is not None:
        for file in os.listdir(dest_dir):
            file = os.path.join(dest_dir, file)
            if not os.path.isfile(file):
                continue

            head_tail = os.path.split(file)
            filename_ext = head_tail[1].split(pdb[0:-4])
            src = os.path.join(file)
            dst = os.path.join(
                head_tail[0], f'{pdb[0:-4]}_state{state}{filename_ext[1]}')

            # check if the file doesn't exist
            if not os.path.exists(dst):
                os.rename(src, dst)

    return analysis, yes



import numpy as np
import matplotlib.patches as mp_patches
import matplotlib.collections as mp_collections
from scipy.spatial.distance import cdist
from rnavigate import styles, data, plots


# Secondary structure diagram ploting functions
def plot_structure_ss(ax, structure, colors):
    """Plot the structure of a secondary structure diagram.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Axes to plot on.
    structure : rnavigate.data.SecondaryStructure
        Secondary structure with diagram drawing coordinates to plot.
    colors : list
        List of colors to use for each nucleotide in the structure.
    """
    x = structure.xcoordinates
    y = structure.ycoordinates

    # ax.scatter(x, y, marker=".", c=colors, **styles.settings['ss']['points'])

    path = mp_patches.Path(np.column_stack([x, y]))
    verts = path.interpolated(steps=2).vertices
    x, y = verts[:, 0], verts[:, 1]
    colors = [x for c in colors for x in [c, c]][1:-1]
    points = np.array([x, y]).T.reshape(-1, 1, 2)
    segments = np.concatenate([points[:-1], points[1:]], axis=1)

    current_color = None
    new_segments = []
    new_colors = []
    for segment, color in zip(segments, colors):
        if current_color is None:
            current_color = color
            current_segment = list(segment)
        elif current_color == color:
            current_segment.append(segment[-1])
        else:
            new_segments.append(current_segment)
            new_colors.append(current_color)
            current_color = color
            current_segment = list(segment)
    new_segments.append(current_segment)
    new_colors.append(current_color)

    lc = mp_collections.LineCollection(
        new_segments, colors=new_colors, **styles.settings["ss"]["structure"]
    )
    ax.add_collection(lc)


def plot_basepairs_ss(ax, structure, bp_style):
    """Plot the basepairs of a secondary structure diagram.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Axes to plot on.
    structure : rnavigate.data.SecondaryStructure
        Secondary structure with diagram drawing coordinates to plot.
    bp_style : "conventional", "dotted" or "line"
        Style of basepairs to plot.
    """
    if bp_style not in ["conventional", "dotted", "line"]:
        raise ValueError('bp_style must be one of "conventional", "dotted" or "line".')
    x = structure.xcoordinates
    y = structure.ycoordinates
    zorder = styles.settings["ss"]["basepairs"]["zorder"]
    for pair in structure.get_pairs():
        x = structure.xcoordinates[[p - 1 for p in pair]]
        y = structure.ycoordinates[[p - 1 for p in pair]]
        xdist = x[1] - x[0]
        ydist = y[1] - y[0]
        if xdist != 0:
            angle_xy = np.arctan(ydist / xdist)
        elif ydist > 0:
            angle_xy = 1 / 2 * np.pi
        elif ydist < 0:
            angle_xy = 3 / 2 * np.pi
        if xdist < 0:
            angle_xy += np.pi
        x_offset = np.cos(angle_xy) / 3
        y_offset = np.sin(angle_xy) / 3
        x_caps = [x[0] + x_offset, x[1] - x_offset]
        y_caps = [y[0] + y_offset, y[1] - y_offset]
        if bp_style == "dotted":
            x_caps_dist = x_caps[0] - x_caps[1]
            y_caps_dist = y_caps[0] - y_caps[1]
            caps_dist = (x_caps_dist**2 + y_caps_dist**2) ** 0.5
            segs = int(max([1 / 3, caps_dist]) * 3) * 2
            x_dots = [x_caps[0] - x_caps_dist * i / segs for i in range(segs + 1)]
            y_dots = [y_caps[0] - y_caps_dist * i / segs for i in range(segs + 1)]
            ax.scatter(x_dots, y_dots, c="grey", marker=".", s=4, zorder=zorder)
        if bp_style == "line":
            ax.plot(x_caps, y_caps, color="grey", zorder=zorder)
        if bp_style == "conventional":
            nts = "".join([structure.sequence[p - 1] for p in pair]).upper()
            # x_caps = [x[0] + i * xdist / 7 for i in [2, 5]]
            # y_caps = [y[0] + i * ydist / 7 for i in [2, 5]]
            if nts in ["UA", "AU"]:
                ax.plot(
                    x_caps, y_caps, color="grey", zorder=zorder, solid_capstyle="butt"
                )
            if nts in ["GU", "UG", "AG", "GA"]:
                x_center = x[0] + xdist / 2
                y_center = y[0] + ydist / 2
                if nts in ["GU", "UG"]:
                    fc = "white"
                    ec = "grey"
                elif nts in ["AG", "GA"]:
                    fc = "grey"
                    ec = "none"
                ax.scatter(
                    x_center,
                    y_center,
                    zorder=zorder + 1,
                    s=3**2,
                    color="grey",
                    marker="o",
                    fc=fc,
                    ec=ec,
                )
            if nts in ["GC", "CG"]:
                ax.plot(
                    x_caps,
                    y_caps,
                    color="grey",
                    zorder=zorder,
                    linewidth=2,
                    solid_capstyle="butt",
                )
                ax.plot(x_caps, y_caps, color="white", zorder=zorder, linewidth=0.7)


def plot_sequence_ss(ax, structure, colors):
    """Plot the sequence of a secondary structure diagram.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Axes to plot on.
    structure : rnavigate.data.SecondaryStructure
        Secondary structure with diagram drawing coordinates to plot.
    colors : list
        List of colors to use for each nucleotide in the structure.
    """
    if isinstance(structure, data.SequenceCircle):
        x = structure.data["Theta"]
        y = np.full(x.shape, structure.radius)
    else:
        x = structure.xcoordinates
        y = structure.ycoordinates
    if colors is None:
        colors = structure.get_colors("gray")
    for nuc in "GUTACgutac":
        mask = [nt == nuc for nt in structure.sequence]
        xcoords = x[mask]
        ycoords = y[mask]
        marker = "$\mathsf{" + nuc + "}$"
        ax.scatter(
            xcoords,
            ycoords,
            marker=marker,
            c=colors[mask],
            **styles.settings["ss"]["sequence"]
        )


def plot_nucleotides_ss(ax, structure, colors):
    """Plot the nucleotides of a secondary structure diagram.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Axes to plot on.
    structure : rnavigate.data.SecondaryStructure
        Secondary structure with diagram drawing coordinates to plot.
    colors : list
        List of colors to use for each nucleotide in the structure.
    """
    if isinstance(structure, data.SequenceCircle):
        x = structure.data["Theta"]
        y = np.full(x.shape, structure.radius)
    else:
        x = structure.xcoordinates
        y = structure.ycoordinates
    if colors is None:
        colors = structure.get_colors("grey")
    ax.scatter(x, y, c=colors, **styles.settings["ss"]["nucleotides"])


def plot_positions_ss(ax, structure, xticks=20):
    """Plot the positions of a secondary structure diagram.

    Label locations are chosen from a point on a circle around each position that is
    the furthest from any other nucleotides. This sometimes causes tick marks and
    labels to overlap with other plot elements.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Axes to plot on.
    structure : rnavigate.data.SecondaryStructure
        Secondary structure with diagram drawing coordinates to plot.
    xticks : int
        Spacing between position labels.
    """
    xs = structure.xcoordinates
    ys = structure.ycoordinates
    thetas = np.pi / 32 * np.arange(64)
    x_shift = np.sin(thetas)
    y_shift = np.cos(thetas)
    ticks, labels = plots.get_nt_ticks(
        sequence=structure.sequence, region=(1, structure.length), gap=xticks
    )
    for nt, label in zip([t - 1 for t in ticks], labels):
        x_nt = xs[nt]
        y_nt = ys[nt]
        x_pos = xs[nt] + (x_shift * 1.5)
        y_pos = ys[nt] + (y_shift * 1.5)
        not_nt = np.arange(structure.length) != nt
        nt_box = (np.abs(xs - xs[nt]) < 4) & (np.abs(ys - ys[nt]) < 4)
        # print(np.vstack((x_pos, y_pos)).T)
        dists = cdist(
            np.vstack((x_pos, y_pos)).T,
            np.vstack((xs[not_nt & nt_box], ys[not_nt & nt_box])).T,
            "euclidean",
        )
        min_dists = np.min(dists, axis=1)
        which_pos = np.where(min_dists == np.max(min_dists))[0][0]
        x_label = x_pos[which_pos]
        y_label = y_pos[which_pos]
        bbox = dict(boxstyle="round", fc="w", ec="w", pad=0.1)
        ax.plot([x_nt, x_label], [y_nt, y_label], color="k", lw=1)
        ax.text(
            x_label,
            y_label,
            label,
            ha="center",
            va="center",
            bbox=bbox,
            **styles.settings["ss"]["positions"]
        )


def plot_interactions_ss(ax, structure, interactions):
    """Plot the interactions as lines over a secondary structure diagram.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Axes to plot on.
    structure : rnavigate.data.SecondaryStructure
        Secondary structure with diagram drawing coordinates.
    interactions : rnavigate.data.Interactions
        Interactions to plot.
    """
    ij_colors = interactions.get_ij_colors()
    segments = []
    colors = []
    for i, j, color in zip(*ij_colors):
        segments.append(
            [
                [structure.xcoordinates[i - 1], structure.ycoordinates[i - 1]],
                [structure.xcoordinates[j - 1], structure.ycoordinates[j - 1]],
            ]
        )
        colors.append(color)
    ax.add_collection(
        mp_collections.LineCollection(
            segments=segments, colors=colors, **styles.settings["ss"]["interactions"]
        )
    )


def plot_annotation_ss(ax, structure, annotation):
    """Highlight regions or nucleotides of interest on a secondary structure diagram.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Axes to plot on.
    structure : rnavigate.data.SecondaryStructure
        Secondary structure with diagram drawing coordinates.
    annotation : rnavigate.data.Annotation
        Annotation to plot.
    """
    color = annotation.color
    if annotation.annotation_type in ["spans", "primers"]:
        for start, end in annotation:
            if start > end:
                start, end = end, start
            x = structure.xcoordinates[start - 1 : end]
            y = structure.ycoordinates[start - 1 : end]
            if start == end:
                ax.scatter(x, y, color=color, **styles.settings["ss"]["sites"])
            else:
                ax.plot(x, y, color=color, **styles.settings["ss"]["spans"])
    elif annotation.annotation_type in ["sites", "group"]:
        sites = np.array(annotation[:]) - 1
        x = structure.xcoordinates[sites]
        y = structure.ycoordinates[sites]
        if annotation.annotation_type == "group":
            ax.plot(x, y, color=color, **styles.settings["ss"]["spans"])
        ax.scatter(x, y, color=color, **styles.settings["ss"]["sites"])



import pandas as pd
import numpy as np
from types import FunctionType
import xml.etree.ElementTree as xmlet
from rnavigate import data


class Profile(data.Data):
    """A class to represent per-nucleotide data.

    Parameters
    ----------
    input_data : str or pandas.DataFrame
        path to a csv or tab file or a pandas DataFrame
        Table must be 1 row for each nucleotide in the sequence.
        table columns must contain these columns:
            A nucleotide position column labelled "Nucleotide"
            A sequence column labelled "Sequence" with 1 of (A, C, G, U, T) per row
                These will be added to the table if `sequence` is provided.
            A data measurement column labelled "Profile" with a float or integer
                Label may be another name if specified in `metric_defaults`
            Optionally: A measurement error column.
                Label must be specified in `metric_defaults`
            Other columns may be present, and set up using `metric_defaults`.
                See `metric_defaults` for more information.
    read_table_kw : dict, optional
        Keyword arguments to pass to pandas.read_table.
        Defaults to None.
    sequence : rnavigate.Sequence or str, optional
        A sequence to use as the reference sequence.
        This is required if `input_data` does not contain a "Sequence" column.
        Defaults to None.
    metric : str, defaults to "default"
        The name of the set of value-to-color options to use.
        "default" specifies:
            "Profile" column is used
            No error rates are present
            Values are normalized to the range [0, 1]
            Values are mapped to colors using the "viridis" colormap
        "Distance" specifies:
            (3-D) "Distance" column is used
            No error rates are present
            Values in the range [5, 50] are normalized to the range [0, 1]
            Values are mapped to colors using the "cool" colormap
        Other options may be defined in `metric_defaults`.
    metric_defaults : dict, optional
        Keys are metric names, to be used with `metric`.
        Values are dictionaries of plotting parameters:
            "metric_column" : str
                The name of the column to use as the metric.
                Plots and analyses that use per-nucleotide data will use this column.
                If "color_column" is not provided, this column also defines colors.
            "error_column" : str or None
                The name of the column to use as the error.
                If None, no error bars are plotted.
            "color_column" : str or None
                The name of the column to use for coloring.
                If None, colors are defined by "metric_column".
            "cmap" : str or list
                The name of the colormap to use.
                If a list, the list of colors to use.
            "normalization" : str
                The type of normalization to use.
                In order to be used with colormaps, values are normalized to either
                be integers for categorical colormaps, or floats in the range [0, 1]
                for continuous colormaps.
                "none" : no normalization is performed
                "min_max" : values are scaled to floats in the range [0, 1] based on
                    the upper and lower bounds defined in "values"
                "0_1" : values are scaled to floats in the range [0, 1] based on
                    the minimum and maximum values in the data
                "bins" : values are scaled an integer based on bins defined by the
                    list of bounds defined in "values"
                "percentiles" : values are scaled to floats in the range [0, 1]
                    based on upper and lower percentile bounds defined by "values"
            "values" : list or None
                The values to use when normalizing the data.
                if "normalization" is "min_max", this should be a list of two values
                    defining the upper and lower bounds.
                if "normalization" is "bins", this should be a list of values
                    of length 1 less than the length of cmap.
                    example: [5, 10, 20] defines 4 bins:
                        (-infinity, 5), [5, 10), [10, 20), [20, infinity)
                if "normalization" is "percentiles", this should be a list of two
                    values defining the upper and lower percentile bounds.
                if "normalization" is "0_1" or "none", this should be None.
            "title" : str, defaults to ""
                The title of the colorbar.
            "ticks" : list, defaults to None
                The tick locations to use for the colorbar. If None, values are
                determined automatically.
            "tick_labels" : list, defaults to None
                The labels to use for the colorbar ticks. If None, values are
                determined automatically from "ticks".
            "extend" : "neither", "both", "min", or "max", defaults to "neither"
                Which ends of the colorbar to extend (places an arrow head).
        Defaults to None.
    name : str, optional
        A name for the data set. Defaults to None.

    Attributes
    ----------
    data : pandas.DataFrame
        The data table
    """

    def __init__(
        self,
        input_data,
        metric="default",
        metric_defaults=None,
        read_table_kw=None,
        sequence=None,
        name=None,
    ):
        """Initialize the Profile object."""
        if metric_defaults is None:
            metric_defaults = {}
        super().__init__(
            input_data=input_data,
            sequence=sequence,
            metric=metric,
            metric_defaults=metric_defaults,
            read_table_kw=read_table_kw,
            name=name,
        )

    @classmethod
    def from_array(cls, input_data, sequence, **kwargs):
        """Construct a Profile object from an array of values.

        Parameters
        ----------
        input_data : list or np.array
            A list or array of values to use as the metric.
        sequence : str
            The RNA sequence.
        **kwargs
            Additional keyword arguments to pass to the Profile constructor.

        Returns
        -------
        Profile
            A Profile object with the provided values.
        """
        sequence = data.Sequence(sequence)
        if len(input_data) != sequence.length:
            message = (
                "RNAVIGATE ERROR\n"
                "===============\n"
                "\tinput array must be the same size as the input sequence\n"
            )
            print(message)
            raise ValueError(message)
        if not all(isinstance(value, (int, float)) for value in input_data):
            message = (
                "RNAVIGATE ERROR\n"
                "===============\n"
                "\tinput array must be all float or integer values\n"
            )
            print(message)
            raise ValueError(message)
        df = pd.DataFrame(
            {
                "Nucleotide": [i for i in range(sequence.length)],
                "Sequence": list(sequence.sequence),
                "Profile": input_data,
            }
        )
        return cls(input_data=df, sequence=sequence, **kwargs)

    @property
    def recreation_kwargs(self):
        """A dictionary of keyword arguments to pass when recreating the object."""
        return {}

    def normalize_sequence(self, t_or_u="U", uppercase=True):
        """Changes the values in self.data["Sequence"] to the normalized sequence.

        Parameters
        ----------
        t_or_u : "T" or "U", Defaults to "U".
            Whether to replace T with U or U with T.
        uppercase : bool, Defaults to True.
            Whether to convert the sequence to uppercase.
        """
        super().normalize_sequence(t_or_u=t_or_u, uppercase=uppercase)
        self.data["Sequence"] = list(self.sequence)

    def get_aligned_data(self, alignment):
        """Returns a new Profile object with the data aligned to a sequence.

        Parameters
        ----------
        alignment : rnavigate.data.SequenceAlignment
            The alignment to use to map rows of self.data to a new sequence.

        Returns
        -------
        Profile
            A new Profile object with the data aligned to the sequence in the
            alignment.
        """
        dataframe = alignment.map_nucleotide_dataframe(self.data)
        return self.__class__(
            input_data=dataframe,
            metric=self._metric,
            metric_defaults=self.metric_defaults,
            sequence=alignment.target_sequence,
            name=self.name,
            **self.recreation_kwargs,
        )

    def copy(self):
        """Returns a copy of the Profile."""
        return self.get_aligned_data(self.null_alignment)

    def get_plotting_dataframe(self):
        """Returns a dataframe with the data to be plotted.

        Returns
        -------
        pandas.DataFrame
            A dataframe with the columns "Nucleotide", "Values", "Errors", and
            "Colors".
        """
        new_names = ["Nucleotide"]
        old_names = ["Nucleotide"]
        old_names.append(self.metric)
        new_names.append("Values")
        if self.error_column is not None:
            old_names.append(self.error_column)
            new_names.append("Errors")
        plotting_dataframe = self.data[old_names].copy()
        plotting_dataframe.columns = new_names
        plotting_dataframe["Colors"] = self.colors
        return plotting_dataframe

    def calculate_windows(
        self,
        column,
        window,
        method="median",
        new_name=None,
        minimum_points=None,
        mask_na=True,
    ):
        """calculates a windowed operation over a column of data.

        Result is stored in a new column. Value of each window is assigned to
        the center position of the window.

        Parameters
        ----------
        column : str
            name of column to perform operation on
        window : int
            window size, must be an odd number
        method : string or function, defaults to "median"
            operation to perform over windows.
            if string, must be "median", "mean", "minimum", or "maximum"
            if function, must take a 1D numpy array as input and return a scalar
        new_name : str, defaults to f"{method}_{window}_nt"
            name of new column for stored result.
        minimum_points : int, defaults to value of `window`
            minimum number of points within each window.
        mask_na : bool, defaults to True
            whether to mask the result of the operation where the original
            column has a nan value.
        """
        if window % 2 != 1:
            raise ValueError("`window` argument must be an odd number.")
        if new_name is None:
            new_name = f"{method}_{window}_nt"
        if minimum_points is None:
            minimum_points = window
        windows = self.data[column].rolling(
            window=window, center=True, min_periods=minimum_points
        )
        if method == "median":
            self.data[new_name] = windows.median()
        elif method == "mean":
            self.data[new_name] = windows.mean()
        elif method == "minimum":
            self.data[new_name] = windows.min()
        elif method == "maximum":
            self.data[new_name] = windows.max()
        elif isinstance(method, FunctionType):
            self.data[new_name] = windows.apply(method)
        else:
            raise ValueError("method argument must be median, mean, maximum or minimum")
        if mask_na:
            self.data.loc[self.data[column].isna(), new_name] = np.nan

    def calculate_gini_index(self, values):
        """Calculate the Gini index of an array of values."""
        # Mean absolute difference
        mad = np.abs(np.subtract.outer(values, values)).mean()
        # Relative mean absolute difference
        rmad = mad / np.mean(values)
        # Gini coefficient
        g = 0.5 * rmad
        return g

    def normalize(
        self,
        profile_column=None,
        new_profile=None,
        error_column=None,
        new_error=None,
        norm_method=None,
        nt_groups=None,
        profile_factors=None,
        **norm_kwargs,
    ):
        """Normalize values in a column, and store in a new column.

        By default, performs ShapeMapper2 boxplot normalization on self.metric
        and stores the result as "Norm_profile".

        Parameters
        ----------
        profile_column : string, defaults to self.metric
            column name of values to normalize
        new_profile : string, defaults to "Norm_profile"
            column name of new normalized values
        error_column : string, defaults to self.error_column
            column name of error values to propagate
        new_error : string, defaults to "Norm_error"
            column name of new propagated error values
        norm_method : string, defaults to "boxplot"
            normalization method to use.
            "DMS" uses self.norm_percentile and nt_groups=['AC', 'UG']
                scales the median of 90th to 95th percentiles to 1
                As and Cs are normalized seperately from Us and Gs
            "eDMS" uses self.norm_eDMS and  nt_groups=['A', 'U', 'C', 'G']
                Applies the new eDMS-MaP normalization.
                Each nucleotide is normalized seperately.
            "boxplot" uses self.norm_boxplot and nt_groups=['AUCG']
                removes outliers (> 1.5 iqr) and scales median to 1
                scales nucleotides together unless specified with nt_groups
            "percentile" uses self.norm_percentile and nt_groups=['AUCG']
                scales the median of 90th to 95th percentiles to 1
                scales nucleotides together unless specified with nt_groups
            Defaults to "boxplot": the default normalization of ShapeMapper
        nt_groups : list of strings, defaults to None
            A list of nucleotides to group
            e.g. ['AUCG'] groups all nts together
                    ['AC', 'UG'] groups As with Cs and Us with Gs
                    ['A', 'C', 'U', 'G'] scales each nt seperately
            Default depends on norm_method
        profile_factors : dictionary, defaults to None
            a scaling factor (float) for each nucleotide. keys must be:
                'A', 'C', 'U', 'G'
            Note: using this argument overrides any calculation of scaling
            Defaults to None
        **norm_kwargs
            these are passed to the norm_method function

        Returns
        -------
        profile_factors : dict
            the new profile scaling factors dictionary
        """
        if profile_column is None:
            profile_column = self.metric
        if new_profile is None:
            new_profile = "Norm_profile"
        if error_column is None:
            error_column = self.error_column
        if new_error is None:
            new_error = "Norm_error"
        profile = self.data[profile_column]
        norm_sequence = self.sequence.upper().replace("T", "U")
        # initialize the error arrays
        if error_column is not None:
            error = self.data[error_column]
            normerr = np.zeros(error.shape)
            mask = profile != 0
            normerr[mask] = (error[mask] / profile[mask]) ** 2
        else:
            normerr = None
        # calculate normalization factors, if appropriate
        if profile_factors is None:
            profile_factors = {nt: np.nan for nt in "ACGU"}
            error_factors = {nt: np.nan for nt in "ACGU"}
            methods_groups = {
                "DMS": (self.norm_percentiles, ["AC", "GU"]),
                "eDMS": (self.norm_eDMS, ["A", "C", "G", "U"]),
                "boxplot": (self.norm_boxplot, nt_groups),
                "percentiles": (self.norm_percentiles, nt_groups),
            }
            norm_method, nt_groups = methods_groups[norm_method]
            if nt_groups is None:
                nt_groups = ["AUCG"]
            for nt_group in nt_groups:
                in_group = [nt in nt_group for nt in norm_sequence]
                prof, err = norm_method(profile[in_group], **norm_kwargs)
                for nt in nt_group:
                    profile_factors[nt] = prof
                    error_factors[nt] = err
        # calculate the new profile
        norm_profile = profile / [profile_factors[nt] for nt in norm_sequence]
        self.data[new_profile] = norm_profile
        # calculate the new errors, if appropriate
        if normerr is not None and error_factors is not None:
            for i in "AUCG":
                mask = [nt == i for nt in norm_sequence]
                normerr[mask] += (error_factors[i] * profile_factors[i]) ** 2
                normerr[mask] = np.sqrt(normerr[mask])
                normerr[mask] *= np.abs(norm_profile[mask])
            self.data[new_error] = normerr
        return profile_factors

    def winsorize(self, column, lower_bound=None, upper_bound=None):
        """Winsorize the data between bounds.

        If either bound is set to None, one-sided Winsorization is performed.

        Parameters
        ----------
        column : string
            the column of data to be winsorized
        lower_bound : Number or None, defaults to None
            Data below this value is set to this value.
            If None, no lower bound is applied.
        upper_bound : Number or None, defaults to None
            Data above this value is set to this value.
            If None, no upper bound is applied.
        """
        if lower_bound is not None:
            below_idx = self.data[column] < lower_bound
            self.data.loc[below_idx, column] = lower_bound
        if upper_bound is not None:
            above_idx = self.data[column] > upper_bound
            self.data.loc[above_idx, column] = upper_bound

    def normalize_external(self, profiles, **kwargs):
        """normalize reactivities using other profiles to normfactors.

        Parameters
        ----------
        profiles : list of rnavigate.data.Profile
            a list of other profiles used to compute scaling factors

        Returns
        -------
        profile_factors : dict
            the new profile scaling factors dictionary
        """
        combined_df = pd.concat([profile.data for profile in profiles])
        combined_profile = Profile(input_data=combined_df)
        nfacs = combined_profile.normalize(**kwargs)
        self.normalize(profile_factors=nfacs)
        return nfacs

    def norm_boxplot(self, values):
        """removes outliers (> 1.5 * IQR) and scales the mean to 1.

        NOTE: This method varies slightly from normalization method used in the
        SHAPEMapper pipeline. Shapemapper sets undefined values to 0, and then
        uses these values when computing iqr and 90th percentile. Including
        these values can skew these result. This method excludes such nan
        values. Other elements are the same.

        Parameters
        ----------
        values : 1D numpy array
            values to normalize

        Returns
        -------
        (float, float)
            scaling factor and error propagation factor
        """
        finite_values = values[np.isfinite(values)]
        p25, p75, p90, p95 = np.percentile(finite_values, [25, 75, 90, 95])
        iqr = p75 - p25
        # filter by iqr
        iqr_values = finite_values[finite_values < (1.5 * iqr + p75)]
        num_finite = len(finite_values)
        num_iqr = len(iqr_values)
        # see if too many values are classified as outliers
        if num_finite < 100 and num_iqr / num_finite < 0.95:
            iqr_values = finite_values[finite_values <= p95]
        elif num_finite >= 100 and num_iqr / num_finite < 0.9:
            iqr_values = finite_values[finite_values < p90]
        new_p90 = np.percentile(iqr_values, 90)
        final_values = iqr_values[iqr_values > new_p90]
        factor = np.mean(final_values)
        error_factor = np.std(final_values) / np.sqrt(len(final_values))
        return factor, error_factor

    def norm_eDMS(self, values):
        """Calculates norm factors following eDMS pernt scheme in ShapeMapper 2.2

        Parameters
        ----------
        values : 1D numpy array
            values to normalize

        Returns
        -------
        (float, float)
            scaling factor and error propagation factor
        """
        # if too few values points, don't normalize
        if len(values) < 10:
            return np.nan, np.nan
        bounds = np.percentile(values, [90.0, 95.0])
        mask = (values >= bounds[0]) & (values < bounds[1])
        normset = values[mask]
        # compute the norm the standard way
        n1 = np.mean(normset)
        try:
            # compute the norm only considering reactive nts
            n2 = np.percentile(values[values > 0.001], 75.0)
        except IndexError:
            n2 = 0
        factor = max(n1, n2)
        # if signal too low, don't norm the values
        if factor < 0.002:
            return np.nan, np.nan
        error_factor = np.std(normset) / np.sqrt(len(normset))
        return factor, error_factor

    def norm_percentiles(
        self, values, lower_bound=90, upper_bound=99, median_or_mean="mean"
    ):
        """Calculates factors to scale the median between percentile bounds to 1.

        Parameters
        ----------
        values : 1D numpy array
            values to normalize
        lower_bound : int or float, optional
            percentile of lower bound, Defaults to 90
        upper_bound : int or float, optional
            percentile of upper bound, Defaults to 99
        median_or_mean : string, optional
            whether to use the median or mean of the values between the bounds.

        Returns
        -------
        (float, float)
            scaling factor and error propagation factor
        """
        if median_or_mean == "mean":
            factor_function = np.mean
        elif median_or_mean == "median":
            factor_function = np.median
        finite_values = values[np.isfinite(values)]
        bounds = np.percentile(finite_values, [lower_bound, upper_bound])
        mask = (finite_values >= bounds[0]) & (finite_values <= bounds[1])
        normset = finite_values[mask]
        factor = factor_function(normset)
        error_factor = np.std(normset) / np.sqrt(len(normset))
        return factor, error_factor


class SHAPEMaP(Profile):
    """A class to represent per-nucleotide SHAPE-MaP data.

    Parameters
    ----------
    input_data : str or pandas.DataFrame
        path to a ShapeMapper2 profile.txt or .map file or a pandas DataFrame
    normalize : "DMS", "eDMS", "boxplot", "percentiles", or None, defaults to None
        The normalization method to use.
        "DMS" uses self.norm_percentile and nt_groups=['AC', 'UG']
            scales the median of 90th to 95th percentiles to 1
            As and Cs are normalized seperately from Us and Gs
        "eDMS" uses self.norm_eDMS and  nt_groups=['A', 'U', 'C', 'G']
            Applies the new eDMS-MaP normalization.
            Each nucleotide is normalized seperately.
        "boxplot" uses self.norm_boxplot and nt_groups=['AUCG']
            removes outliers (> 1.5 iqr) and scales median to 1
            scales nucleotides together unless specified with nt_groups
        "percentiles" uses self.norm_percentile and nt_groups=['AUCG']
            scales the median of 90th to 95th percentiles to 1
            scales nucleotides together unless specified with nt_groups
        Defaults to None: no normalization is performed
    read_table_kw : dict, optional
        Keyword arguments to pass to pandas.read_table. These are not necessary for
        profile.txt and .map files.
        Defaults to None.
    sequence : rnavigate.Sequence or str, optional
        A sequence to use as the reference sequence. This is not necessary for
        profile.txt and .map files.
        Defaults to None.
    metric : str, defaults to "Norm_profile"
        The name of the set of value-to-color options to use.
        "Norm_profile" specifies:
            "Norm_profile" column is used
            "Norm_stderr" column is used for error bars
            Values are normalized to bins:
                (-inf, -0.4), [-0.4, 0.4), [0.4, 0.85), [0.85, 2), [2, inf)
            Bins are mapped to "grey", "black", "orange", "red", "red"
        Other options may be defined in `metric_defaults`.
    metric_defaults : dict, optional
        Keys are metric names, to be used with `metric`.
        Values are dictionaries of plotting parameters:
            "metric_column" : str
                The name of the column to use as the metric.
                Plots and analyses that use per-nucleotide data will use this column.
                If "color_column" is not provided, this column also defines colors.
            "error_column" : str or None
                The name of the column to use as the error.
                If None, no error bars are plotted.
            "color_column" : str or None
                The name of the column to use for coloring.
                If None, colors are defined by "metric_column".
            "cmap" : str or list
                The name of the colormap to use.
                If a list, the list of colors to use.
            "normalization" : str
                The type of normalization to use.
                In order to be used with colormaps, values are normalized to either
                be integers for categorical colormaps, or floats in the range [0, 1]
                for continuous colormaps.
                "none" : no normalization is performed
                "min_max" : values are scaled to floats in the range [0, 1] based on
                    the upper and lower bounds defined in "values"
                "0_1" : values are scaled to floats in the range [0, 1] based on
                    the minimum and maximum values in the data
                "bins" : values are scaled an integer based on bins defined by the
                    list of bounds defined in "values"
                "percentiles" : values are scaled to floats in the range [0, 1]
                    based on upper and lower percentile bounds defined by "values"
            "values" : list or None
                The values to use when normalizing the data.
                if "normalization" is "min_max", this should be a list of two values
                    defining the upper and lower bounds.
                if "normalization" is "bins", this should be a list of values
                    of length 1 less than the length of cmap.
                    example: [5, 10, 20] defines 4 bins:
                        (-infinity, 5), [5, 10), [10, 20), [20, infinity)
                if "normalization" is "percentiles", this should be a list of two
                    values defining the upper and lower percentile bounds.
                if "normalization" is "0_1" or "none", this should be None.
            "title" : str, defaults to ""
                The title of the colorbar.
            "ticks" : list, defaults to None
                The tick locations to use for the colorbar. If None, values are
                determined automatically.
            "tick_labels" : list, defaults to None
                The labels to use for the colorbar ticks. If None, values are
                determined automatically from "ticks".
            "extend" : "neither", "both", "min", or "max", defaults to "neither"
                Which ends of the colorbar to extend (places an arrow head).
        Defaults to None.
    log : str, optional
        Path to a ShapeMapper v2 shapemap_log.txt file with mutations-per-molecule
        and read-length histograms. These will be present if the --per-read-histogram
        flag was used when running ShapeMapper v2.
        Currently, this is not working with ShapeMapper v2.2 files.
        Defaults to None.
    name : str, optional
        A name for the data set. Defaults to None.

    Attributes
    ----------
    data : pandas.DataFrame
        The data table
    """

    def __init__(
        self,
        input_data,
        normalize=None,
        read_table_kw=None,
        sequence=None,
        metric="Norm_profile",
        metric_defaults=None,
        log=None,
        name=None,
    ):
        """Initialize the SHAPEMaP object."""
        self.read_lengths, self.mutations_per_molecule = self.read_log(log)
        if metric_defaults is None:
            metric_defaults = {}
        metric_defaults = {
            "Norm_profile": {
                "metric_column": "Norm_profile",
                "error_column": "Norm_stderr",
                "cmap": ["grey", "black", "orange", "red", "red"],
                "normalization": "bins",
                "values": [-0.4, 0.4, 0.85, 2],
                "title": "SHAPE Reactivity",
                "extend": "both",
            }
        } | metric_defaults
        if (
            isinstance(input_data, str)
            and input_data.endswith(".map")
            and read_table_kw is None
        ):
            read_table_kw = {
                "names": ["Nucleotide", "Norm_profile", "Norm_stderr", "Sequence"],
                "na_values": "-999",
            }

        super().__init__(
            input_data=input_data,
            read_table_kw=read_table_kw,
            sequence=sequence,
            metric=metric,
            metric_defaults=metric_defaults,
            name=name,
        )
        if normalize is not None:
            self.normalize(
                profile_column="HQ_profile",
                new_profile="Norm_profile",
                error_column="HQ_stderr",
                new_error="Norm_stderr",
                norm_method=normalize,
            )

    @classmethod
    def from_rnaframework(cls, input_data, normalize=None):
        """Construct a SHAPEMaP object from an RNAFramework output file.

        Parameters
        ----------
        input_data : str
            path to an RNAFramework .xml reactivities file
        normalize : "DMS", "eDMS", "boxplot", "percentiles", or None, defaults to None
            The normalization method to use.
            "DMS" uses self.norm_percentile and nt_groups=['AC', 'UG']
                scales the median of 90th to 95th percentiles to 1
                As and Cs are normalized seperately from Us and Gs
            "eDMS" uses self.norm_eDMS and  nt_groups=['A', 'U', 'C', 'G']
                Applies the new eDMS-MaP normalization.
                Each nucleotide is normalized seperately.
            "boxplot" uses self.norm_boxplot and nt_groups=['AUCG']
                removes outliers (> 1.5 iqr) and scales median to 1
                scales nucleotides together unless specified with nt_groups
            "percentiles" uses self.norm_percentile and nt_groups=['AUCG']
                scales the median of 90th to 95th percentiles to 1
                scales nucleotides together unless specified with nt_groups
            Defaults to None: no normalization is performed

        Returns
        -------
        SHAPEMaP
            A SHAPEMaP object with the provided values.
        """
        tree = xmlet.parse(input_data)
        root = tree.getroot()
        sequence = root.find("./transcript/sequence").text
        sequence = [nt for nt in sequence if nt in "ATCG"]
        reactivities = root.find("./transcript/reactivity").text
        reactivities = [float(rx) for rx in reactivities.split(",")]
        input_data = pd.DataFrame(
            {
                "Nucleotide": [n + 1 for n in range(len(sequence))],
                "Sequence": sequence,
                "Norm_profile": reactivities,
            }
        )
        profile = cls(input_data=input_data, normalize=normalize)
        return profile

    def read_log(self, log):
        """Read the ShapeMapper log file.

        Parameters
        ----------
        log : str
            Path to a ShapeMapper v2 shapemap_log.txt file with
            mutations-per-molecule and read-length histograms.

        Returns
        -------
        read_lengths : pandas.DataFrame
            A dataframe with the columns "Read_length", "Modified_read_length",
            and "Untreated_read_length".
        mutations_per_molecule : pandas.DataFrame
            A dataframe with the columns "Mutation_count",
            "Modified_mutations_per_molecule", and
            "Untreated_mutations_per_molecule".
        """
        if log is None:
            return None, None
        with open(log, "r") as f:
            flist = list(f)
            log_format_test = 0
            for i, line in enumerate(flist):
                if line.startswith("  |MutationCounter_Modified"):
                    log_format_test += 1
                    modlength = []
                    for x in flist[i + 6 : i + 27]:
                        modlength.append(float(x.strip().split("\t")[1]))
                    modmuts = []
                    for x in flist[i + 32 : i + 53]:
                        modmuts.append(float(x.strip().split("\t")[1]))
                if line.startswith("  |MutationCounter_Untreated"):
                    log_format_test += 1
                    untlength = []
                    for x in flist[i + 6 : i + 27]:
                        untlength.append(float(x.strip().split("\t")[1]))
                    untmuts = []
                    for x in flist[i + 32 : i + 53]:
                        untmuts.append(float(x.strip().split("\t")[1]))
        message = (
            "Histogram data missing from log file. Requires"
            + " --per-read-histogram flag when running ShapeMapper."
        )
        assert log_format_test >= 2, message
        read_lengths = pd.DataFrame(
            {
                "Read_length": [
                    "0-49",
                    "50-99",
                    "100-149",
                    "150-199",
                    "200-249",
                    "250-299",
                    "300-349",
                    "350-399",
                    "400-449",
                    "450-499",
                    "500-549",
                    "550-599",
                    "600-649",
                    "650-699",
                    "700-749",
                    "750-799",
                    "800-849",
                    "850-899",
                    "900-949",
                    "950-999",
                    ">1000",
                ],
                "Modified_read_length": modlength,
                "Untreated_read_length": untlength,
            }
        )
        mutations_per_molecule = pd.DataFrame(
            {
                "Mutation_count": np.arange(21),
                "Modified_mutations_per_molecule": modmuts,
                "Untreated_mutations_per_molecule": untmuts,
            }
        )
        return read_lengths, mutations_per_molecule


class DanceMaP(SHAPEMaP):
    """A class to represent per-nucleotide DanceMaP data.

    Parameters
    ----------
    input_data : str or pandas.DataFrame
        path to a DanceMapper reactivities.txt file or a pandas DataFrame
    component : int
        Which component of the DanceMapper ensemble to read in (0-indexed).
    read_table_kw : dict, optional
        Keyword arguments to pass to pandas.read_table. These are not necessary for
        reactivities.txt files.
        Defaults to None.
    sequence : rnavigate.Sequence or str, optional
        A sequence to use as the reference sequence. This is not necessary for
        reactivities.txt files.
        Defaults to None.
    metric : str, defaults to "Norm_profile"
        The name of the set of value-to-color options to use.
    """

    def __init__(
        self,
        input_data,
        component,
        read_table_kw=None,
        sequence=None,
        metric="Norm_profile",
        metric_defaults=None,
        name=None,
    ):
        self.component = component
        super().__init__(
            input_data=input_data,
            read_table_kw=read_table_kw,
            sequence=sequence,
            metric=metric,
            metric_defaults=metric_defaults,
            name=name,
        )

    @property
    def recreation_kwargs(self):
        return {"component": self.component}

    def read_file(self, input_data, read_table_kw={}):
        # parse header
        self.filepath = input_data
        with open(self.filepath) as f:
            header1 = f.readline()
            header2 = f.readline()
        self.header = header1 + header2
        self.components = int(header1.strip().split()[0])
        self.percents = [float(x) for x in header2.strip().split()[1:]]
        self.percent = self.percents[self.component]
        # parse datatable
        read_table_kw["names"] = [
            "Nucleotide",
            "Sequence",
            "Norm_profile",
            "Modified_rate",
            "Untreated_rate",
        ]
        col_offset = 3 * self.component
        bg_col = 3 * self.components + 2
        read_table_kw["usecols"] = [0, 1, 2 + col_offset, 3 + col_offset, bg_col]
        df = pd.read_table(self.filepath, header=2, **read_table_kw)
        # some rows have an "i" added to the final column
        stripped = []
        for x in df["Untreated_rate"]:
            if type(x) is float:
                stripped.append(x)
            else:
                stripped.append(float(x.rstrip(" i")))
        df["Untreated_rate"] = stripped
        df = df.eval("Reactivity_profile = Modified_rate - Untreated_rate")
        return df


class RNPMaP(Profile):
    def __init__(
        self,
        input_data,
        read_table_kw=None,
        sequence=None,
        metric="NormedP",
        metric_defaults=None,
        name=None,
    ):
        if metric_defaults is None:
            metric_defaults = {}
        if read_table_kw is None:
            read_table_kw = {}
        metric_defaults = {
            "NormedP": {
                "metric_column": "NormedP",
                "color_column": "RNPsite",
                "cmap": ["silver", "limegreen"],
                "normalization": "none",
                "values": None,
            }
        } | metric_defaults
        read_table_kw = {"sep": ","} | read_table_kw
        super().__init__(
            input_data=input_data,
            read_table_kw=read_table_kw,
            sequence=sequence,
            metric=metric,
            metric_defaults=metric_defaults,
            name=name,
        )


class DeltaProfile(Profile):
    def __init__(
        self,
        profile1,
        profile2,
        metric=None,
        metric_defaults=None,
        name=None,
    ):
        if metric is None:
            metric = profile1.metric

        columns = ["Nucleotide", "Sequence", metric]
        alignment = data.SequenceAlignment(profile2, profile1)
        profile2 = profile2.get_aligned_data(alignment)
        new_data = pd.merge(
            profile1.data[columns],
            profile2.data[columns],
            how="left",
            on=["Nucleotide"],
            suffixes=["_1", "_2"],
        )
        new_data.eval(f"Delta_profile = {metric}_1 - {metric}_2", inplace=True)
        metric_defaults = {
            "Delta_profile": {
                "metric_column": "Delta_profile",
                "error_column": None,
                "color_column": None,
                "cmap": "coolwarm",
                "normalization": "min_max",
                "values": [-0.8, 0.8],
            }
        } | metric_defaults
        super().__init__(
            input_data=new_data,
            sequence=profile1.sequence,
            metric="Delta_profile",
            metric_defaults=metric_defaults,
            name=name,
        )



from rnavigate.transcriptomics.bed import BedFile, NarrowPeak
from rnavigate.transcriptomics.transcriptome import (
    Transcript,
    Transcriptome,
)
from rnavigate.transcriptomics.eclip import (
    eCLIPDatabase,
    download_eclip_peaks,
)

__all__ = [
    "BedFile",
    "NarrowPeak",
    "Transcriptome",
    "Transcript",
    "eCLIPDatabase",
    "download_eclip_peaks",
]



import numpy as np
import pandas as pd
import seaborn as sns
from rnavigate import plots


class QC(plots.Plot):
    """Plot QC data from log files.

    Parameters
    ----------
    num_samples : int
        Number of samples to plot.

    Attributes
    ----------
    fig : matplotlib.figure.Figure
        Figure object.
    ax_muts_unt : matplotlib.axes.Axes
        Axes object for the mutations per molecule plot of untreated samples.
    ax_muts_mod : matplotlib.axes.Axes
        Axes object for the mutations per molecule plot of modified samples.
    ax_read_unt : matplotlib.axes.Axes
        Axes object for the read length plot of untreated samples.
    ax_read_mod : matplotlib.axes.Axes
        Axes object for the read length plot of modified samples.
    ax_boxplot : matplotlib.axes.Axes
        Axes object for the boxplot of mutation rates.
    axes : numpy.ndarray of matplotlib.axes.Axes
        Array of axes objects.
    i : int
        Index of the current plot.
    """

    def __init__(self, num_samples):
        """Initialize the plot."""
        super().__init__(num_samples)
        if self.length == 1:
            self.ax_muts_unt = self.axes[0, 0]
            self.ax_muts_mod = self.axes[0, 0]
            self.ax_read_unt = self.axes[0, 1]
            self.ax_read_mod = self.axes[0, 1]
            self.ax_boxplot = self.axes[0, 2]
        elif self.length > 1:
            self.ax_muts_unt = self.axes[0, 0]
            self.ax_muts_mod = self.axes[0, 1]
            self.ax_muts_mod.get_shared_x_axes().join(
                self.ax_muts_mod, self.ax_muts_unt
            )
            self.ax_muts_mod.get_shared_y_axes().join(
                self.ax_muts_mod, self.ax_muts_unt
            )
            self.ax_muts_mod.set_yticklabels([])
            self.ax_read_unt = self.axes[0, 2]
            self.ax_read_mod = self.axes[0, 3]
            self.ax_muts_mod.get_shared_x_axes().join(
                self.ax_read_mod, self.ax_read_unt
            )
            self.ax_muts_mod.get_shared_y_axes().join(
                self.ax_read_mod, self.ax_read_unt
            )
            self.ax_read_mod.set_yticklabels([])
            gs = self.axes[1, 0].get_gridspec()
            for ax in self.axes[1, :]:
                ax.remove()
            self.ax_boxplot = self.fig.add_subplot(gs[1, :])
        self.ax_muts_unt.set(
            xlabel="Mutations per molecule",
            ylabel="Percentage of Reads",
            title="Untreated",
        )
        self.ax_muts_mod.set(xlabel="Mutations per molecule", title="Modified")
        self.ax_read_unt.set(
            xticks=range(12),
            xlabel="Read Length",
            ylabel="Percentage of Reads",
            title="Untreated",
        )
        self.ax_read_mod.set(xticks=range(12), xlabel="Read Length", title="Modified")
        if self.length == 1:
            self.ax_muts_mod.set_title("Mutations per Molecule Distribution")
            self.ax_read_mod.set_title("Read Length Distribution")
            self.ax_boxplot.set_title("Mutation rates")
        xticklabels = [f"{x*50}" for x in range(12)]
        self.ax_read_unt.set_xticklabels(xticklabels, rotation=45)
        self.ax_read_mod.set_xticklabels(xticklabels, rotation=45)
        self.profiles = []

    def set_figure_size(
        self,
        height_ax_rel=None,
        width_ax_rel=None,
        width_ax_in=2,
        height_ax_in=2,
        height_gap_in=1,
        width_gap_in=1,
        top_in=1,
        bottom_in=0.5,
        left_in=1,
        right_in=0.5,
    ):
        """Set the size of the figure.

        Parameters
        ----------
        height_ax_rel : float, optional
            Height of the axes relative to the y-axis limits.
        width_ax_rel : float, optional
            Width of the axes relative to the x-axis limits.
        width_ax_in : float, defaults to 2
            Width of the axes in inches.
        height_ax_in : float, defaults to 2
            Height of the axes in inches.
        height_gap_in : float, defaults to 1
            Height of the gap between axes in inches.
        width_gap_in : float, defaults to 1
            Width of the gap between axes in inches.
        top_in : float, defaults to 1
            Height of the top margin in inches.
        bottom_in : float, defaults to 0.5
            Height of the bottom margin in inches.
        left_in : float, defaults to 0.5
            Width of the left margin in inches.
        right_in : float, defaults to 0.5
            Width of the right margin in inches.
        """
        super().set_figure_size(
            height_ax_rel=height_ax_rel,
            width_ax_rel=width_ax_rel,
            width_ax_in=width_ax_in,
            height_ax_in=height_ax_in,
            height_gap_in=height_gap_in,
            width_gap_in=width_gap_in,
            top_in=top_in,
            bottom_in=bottom_in,
            left_in=left_in,
            right_in=right_in,
        )

    def get_rows_columns(self, rows=None, cols=None):
        """Get the number of rows and columns.

        Parameters
        ----------
        rows : int, optional
            Number of rows. This is ignored.
        cols : int, optional
            Number of columns. This is ignored.

        Returns
        -------
        rows : int
            Number of rows. 1 if there is only one sample, 2 otherwise.
        cols : int
            Number of columns. 3 if there is only one sample, 4 otherwise.
        """
        if self.length == 1:
            return (1, 3)
        else:
            return (2, 4)

    def plot_data(self, profile, label):
        """Plot the data.

        Parameters
        ----------
        profile : rnavigate.profile.Profile
            Profile object.
        label : str
            Label for the sample.
        """
        self.plot_mutations_per_molecule(profile=profile, label=label)
        self.plot_read_lengths(profile=profile, label=label)
        self.profiles.append(profile)
        self.i += 1
        if self.i == self.length:
            handles, labels = self.ax_muts_unt.get_legend_handles_labels()
            if self.length == 1:
                labels = ["Modified", "Untreated"]
            self.ax_muts_unt.legend(handles, labels, title="Samples", loc=1)
            if self.length == 1:
                labels = [label]
            self.make_boxplot(labels)

    def plot_mutations_per_molecule(self, profile, label, upper_limit=12):
        """Plot the mutations per molecule.

        Parameters
        ----------
        profile : rnavigate.profile.Profile
            Profile object.
        label : str
            Label for the sample.
        upper_limit : int, optional
            Upper limit of the x-axis.
        """
        df = profile.mutations_per_molecule
        if df is None:
            raise ValueError("profile is missing QC data from log file")
        x = df.loc[:upper_limit, "Mutation_count"]
        y1 = df.loc[:upper_limit, "Modified_mutations_per_molecule"]
        self.ax_muts_mod.plot(x, y1, label=label)
        y2 = df.loc[:upper_limit, "Untreated_mutations_per_molecule"]
        self.ax_muts_unt.plot(x, y2, label=label)

    def plot_read_lengths(self, profile, label, upper_limit=12):
        """Plot the read lengths.

        Parameters
        ----------
        profile : rnavigate.profile.Profile
            Profile object.
        label : str
            Label for the sample.
        upper_limit : int, optional
            Upper limit of the x-axis.
        """
        df = profile.read_lengths
        if self.length == 1:
            width = 0.4
            x1 = np.arange(upper_limit) - 0.6 + width
            x2 = x1 + width
        else:
            width = 0.8 / self.length
            x1 = np.arange(upper_limit) - 0.4 - (width / 2) + (width * self.i)
            x2 = x1
        y1 = df.loc[: upper_limit - 1, "Modified_read_length"]
        self.ax_read_mod.bar(x1, y1, width, label=label)
        y2 = df.loc[: upper_limit - 1, "Untreated_read_length"]
        self.ax_read_unt.bar(x2, y2, width, label=label)

    def make_boxplot(self, labels):
        """Make the boxplot of mutation rates.

        Parameters
        ----------
        labels : list of str
            Labels for the samples.
        """
        cols = ["Modified_rate", "Untreated_rate"]
        data = []
        for i, profile in enumerate(self.profiles):
            data.append(profile.data[cols].copy().assign(Sample=i + 1))
        data = pd.concat(data)
        data = pd.melt(data, id_vars=["Sample"], var_name=["Rate"])
        ax = sns.violinplot(
            x="Sample",
            y="value",
            hue="Rate",
            data=data,
            ax=self.ax_boxplot,
            split=True,
            scale="width",
        )
        ax.set(
            #    yscale='log',
            #    ylim=(0.00003, 0.3),
            ylabel="Mutation Rate",
            xticklabels=labels,
        )



"""Contains PlottingArgumentParser for plotting_functions.py and fit_data for
retreiving aligned data objects"""

from rnavigate import data, Sample
from rnavigate.data_loading import get_sequence


__all__ = ["fit_data", "PlottingArgumentParser"]


def _parse_plot_kwargs(plot_kwargs, plot):
    if plot_kwargs is None:
        plot_kwargs = {}
    elif not isinstance(plot_kwargs, dict):
        raise ValueError(
            "plot_kwargs must be a dictionary of keyword-arguments "
            f"to be passed to {plot}."
        )
    return plot_kwargs


def fit_data(data_object, alignment):
    """Given a sample and list of sample.data keys, Data objects are mapped to
    sequence

    Parameters
    ----------
        data_object : Data object or list or dict of Data objects
            Data object to be mapped to sequence
        alignment : SequenceAlignment
            SequenceAlignment object to be used for mapping

    Returns
    -------
        Data object or list or dict of Data objects
            A new copy of each Data object mapped to a new sequence
    """
    if alignment is None:
        return data_object
    if isinstance(data_object, dict):
        return {k: fit_data(v, alignment) for k, v in data_object.items()}
    if isinstance(data_object, list):
        return [fit_data(v, alignment) for v in data_object]
    if isinstance(data_object, data.Sequence):
        if isinstance(data_object, data.PDB):
            return data_object
        alignment = data.AlignmentChain(
            data.SequenceAlignment(data_object, alignment.starting_sequence), alignment
        )
        return data_object.get_aligned_data(alignment)
    else:
        return data_object


class PlottingArgumentParser:
    """Parse arguments for high-level plotting functions.

    Given samples list and data keywords, returns a dictionary of data objects
    for each sample, optionally aligned to a target sequence.
    """

    def __init__(self, samples, labels, alignment=None, **data_dict):
        self.samples = self._parse_samples(samples)
        self.labels = self._parse_labels(labels)
        self.rows = len(samples)
        self.cols = 1
        # parse special formats
        for key, value in data_dict.items():
            if key == "interactions":
                data_dict[key] = self._parse_interactions(value)
                self.cols = len(data_dict[key])
            elif "interactions" in key:
                data_dict[key] = self._parse_interactions(value, False)
            elif key == "annotations":
                data_dict[key] = self._parse_annotations(value)
        self.data_dicts = []
        classes = {
            # in order of approximate usage
            "profile": data.Profile,
            "annotations": data.Annotation,
            "structure": (data.SecondaryStructure, data.PDB),
            # "interactions": data.Interactions, # _parse_interactions instead
            "domains": list,
        }
        for sample, label in zip(self.samples, self.labels):
            this_data_dict = {"label": label}
            for key, value in data_dict.items():
                if key == "sequence":
                    new_value = get_sequence(value, sample)
                    new_value = fit_data(new_value, alignment)
                    this_data_dict[key] = new_value
                    continue
                if ("interactions" in key) and key != "interactions":
                    sample.filter_interactions(**value)
                    new_value = sample.get_data(
                        value["interactions"], data.Interactions
                    )
                    new_value = fit_data(new_value, alignment)
                    this_data_dict[key] = new_value
                    continue
                for name, data_class in classes.items():
                    if name in key:
                        new_value = sample.get_data(value, data_class)
                        new_value = fit_data(new_value, alignment)
                        this_data_dict[key] = new_value
                        break
                else:
                    this_data_dict[key] = value
            if "interactions" not in data_dict:
                self.data_dicts.append(this_data_dict)
            else:
                for each_interaction in data_dict["interactions"]:
                    sample.filter_interactions(**each_interaction)
                    new_value = each_interaction["interactions"]
                    new_value = sample.get_data(new_value, data.Interactions)
                    new_dict = {"interactions": fit_data(new_value, alignment)}
                    self.data_dicts.append(this_data_dict | new_dict)

    def update_rows_cols(self, plot_kwargs):
        if self.rows > 1 and self.cols > 1:
            plot_kwargs["rows"] = self.rows
            plot_kwargs["cols"] = self.cols

    @property
    def num_samples(self):
        return self.rows * self.cols

    def _parse_samples(self, samples):
        error = ValueError("`samples` must be a list of rnavigate.Sample")
        if isinstance(samples, Sample):
            samples = list(samples)
        elif not isinstance(samples, list):
            raise error
        return samples

    def _parse_labels(self, labels):
        """Ensures the value of labels is a list, same length as samples."""
        error = ValueError(
            "labels must be a list of strings of length equal "
            "to length of sample list."
        )
        if labels is None:
            labels = [sample.sample for sample in self.samples]
        elif not isinstance(labels, list):
            raise error
        elif len(labels) != len(self.samples):
            raise error
        return labels

    def _parse_annotations(self, annotations):
        """Ensures the value of annotations is a list.

        Parameters
        ----------
            annotations : None, list of str or rnav.data.Annotations
                None is converted to []
                a single string or data object is enclosed in a list

        Returns
        -------
            annotations list
                a list of annotations data or data keywords (or an empty list)
        """
        error = ValueError(
            "annotations must be a list containing data keywords or objects"
        )
        if annotations is None:
            return []
        if isinstance(annotations, list):
            for annotation in annotations:
                if not isinstance(annotation, (data.Annotation, str)):
                    raise error
            return annotations
        if isinstance(annotations, (data.Annotation, str)):
            return [annotations]
        raise error

    def _parse_interactions(self, interactions, return_list=True):
        """Ensure interactions is the expected format.

        interactions must follow one of the following formats:
        format 1: a data keyword or data object
        format 2: dictionary containing format 1:
                    {'interactions': format 1, 'filter': True}
        format 3 (if return_list is True): list of format 2 dictionaries:
                    [format 2, format 2]
        1 is converted to 2 if return_list is False.
        1 and 2 are converted to 3 if return_list is True.

        Parameters
        ----------
            interactions : format 1, 2, or 3
            return_list : bool, defaults to True
                Whether to return format 3, otherwise returns format 2

        Returns
        -------
            interactions dict
                format 2 if return_list is False
            interactions list
                format 3 if return_list is True
        """
        error = ValueError(
            "interactions must follow one of the following formats:\n"
            "format 1: a data keyword or data object\n"
            "format 2: dictionary containing format 1:\n"
            "            {'interactions': format 1, 'filter': True}\n"
            "format 3: list of format 2 dictionaries:\n"
            "            [format 2, format 2]\n"
        )
        if interactions is None:
            if return_list:
                return [{"interactions": None}]
            else:
                return {"interactions": None}
        if isinstance(interactions, (data.Interactions, str)):
            interactions = {"interactions": interactions}
        if isinstance(interactions, dict):
            if return_list:
                return [interactions]
            else:
                return interactions
        if isinstance(interactions, list) and return_list:
            return interactions
        raise error
